/*
 * last modified---
 * 	03-22-17/jet/ rev 0.7.1
 * 	08-27-15/jet/ rev 0.6.1, patches for use with OnionPay
 * 	07-17-15/jet/ clear asset/token config on logout; rev 0.6.0
 * 	10-02-14/jet/ SSM -> generic SVM; add m_SVMPlugin
 * 	08-23-14/jet/ add reorderGateways() and its usage, to speed up logins
 * 	08-19-14/jet/ 0.4.4: make JComboBox usage 1.7-compliant; retrofit SV
 * 					  Spark plugin changes; update PageKite gateway list;
 * 					  use ~/.silentvault for files instead of ~/.vouchers
 * 	03-19-14/jet/ 0.4.3: new OFS gateway and PXGOLD asset support; fix bad
 * 					  voucher sig error for certain sigs created with Java-6;
 * 					  add manual entry for OFS gateway, and connect on choice;
 *						  add "Get JID" menu item which copies to clipboard;
 *						  add doDisconnect()
 * 	01-29-14/jet/ when logging in after timeout, auto-reconnect first
 * 	01-14-14/jet/ add handling for -p option to main() for manual invocation
 * 	11-21-13/jet/ require a connect to :80 at the OFS prior to XMPP c2s open;
 * 					  when SDS or DHT connects fail, retry with prior :80 connect
 * 	11-09-13/jet/ 0.3, silence connection debug and remove beta designation,
 * 					  support TestNet property
 * 	10-12-13/jet/ support separate XMPP domain definitions for OFS/SDS/DHT
 * 					  along with multiple host:port paths
 * 	06-13-13/jet/ add doDisco()
 * 	06-05-13/jet/ SilentVault client version
 * 	05-17-13/kbw/ 1.0.11, add SB- asset types, build with Smack 3.3.0
 * 	01-17-13/kbw/ 1.0.10, add voucher redemption support and NPV calculation
 * 	11-12-12/kbw/ 1.0.9, support selection of multiple OFSen (up to 10); also
 * 					  various bugfixes
 * 	06-25-12/kbw/ 1.0.8, Java 7 support
 * 	03-02-12/kbw/ 1.0.7, Smack 3.2.2 support + Bitcoin prices
 * 	01-18-12/kbw/ 1.0.6, support receipt download
 * 	10-12-11/kbw/ 1.0.5, accommodate AssetPriceCache implementing TimerTask
 * 	08-05-11/kbw/ 1.0.4
 * 	06-24-11/kbw/ for 1.0.3, add m_LogDir and getLogDir()
 * 	05-18-11/kbw/ roll revision to release 1.0.0
 * 	04-12-11/kbw/ use separate thread for display in setDefaultStatusBar()
 * 	03-11-11/kbw/ mods for having AssetPriceCache get prices off of the OFS
 * 	12-13-10/kbw/ impose timeout on OFS connect, instead of using sleep
 * 	11-18-10/kbw/ add build date to M_Version
 * 	11-08-10/kbw/ use USERPROFILE as alternate for user.home on Windows
 * 	10-28-10/kbw/ move Split, Merge, Validate into Advanced menu
 * 	10-21-10/kbw/ add register of popup windows to be cleared on logout
 * 	10-13-10/kbw/ move status bar here from HomePane, reduce vertical size
 * 	09-17-10/kbw/ add F10 menu item Help, and subs Screen, Tutorials, About
 * 	09-09-10/kbw/ add window listener to force logout on window close
 * 	08-17-10/kbw/ add L&F specification, default button font
 * 	04-19-10/kbw/ add Advanced menu tab and logout; restore web listener
 * 	03-14-10/kbw/ support CC of logged errors to listener JID from JNLP file
 * 	03-08-10/kbw/ specify OFS host/port via system properties from JNLP file
 * 	03-05-10/kbw/ remove web listener functionality
 * 	03-02-10/kbw/ more stand-alone work; add menu bar
 * 	02-23-10/kbw/ stand-alone version with main(), extending JFrame
 *
 * purpose---
 * 	main class for webapp implementation of Silent-Vault client
 */

package org.silentvault.client;

import javax.swing.*;
import java.awt.GridLayout;
import java.awt.Dimension;
import java.awt.Cursor;
import java.awt.Font;
import java.awt.Color;
import java.awt.Point;
import java.awt.Component;
import java.awt.GraphicsEnvironment;
import java.awt.event.KeyEvent;
import java.awt.event.ActionEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.datatransfer.StringSelection;
import java.awt.datatransfer.Clipboard;
import java.awt.Toolkit;

import org.jivesoftware.resource.SparkRes;

import org.jivesoftware.smack.provider.*;
import org.jivesoftware.smack.XMPPConnection;
import org.jivesoftware.smack.XMPPException;
import org.jivesoftware.smack.ConnectionConfiguration;
import org.jivesoftware.smack.ConnectionListener;
import org.jivesoftware.smack.SmackConfiguration;
import org.jivesoftware.smack.packet.IQ;
import org.jivesoftware.smack.packet.Packet;
import org.jivesoftware.smack.packet.PacketExtension;
import org.jivesoftware.smack.packet.Message;
import org.jivesoftware.smack.PacketCollector;
import org.jivesoftware.smack.filter.PacketFilter;
import org.jivesoftware.smack.filter.PacketIDFilter;
import org.jivesoftware.smackx.packet.DiscoverItems;

import org.silentvault.client.ui.*;
import org.silentvault.client.ui.wallet.*;
import org.silentvault.client.ui.svx.*;
import org.silentvault.client.ui.svm.*;

import java.security.*;
import java.util.*;
import java.net.InetSocketAddress;
import java.net.URL;
import java.net.URLConnection;
import java.net.MalformedURLException;
import java.net.HttpURLConnection;
import java.net.ProtocolException;
import java.io.File;
import java.io.IOException;
import java.io.FileInputStream;
import java.io.FileNotFoundException;


/**
 * Implements the SilentVault wallet client as a WebApp (web start) program.
 * This allows the application to act as a voucher safe client
 * within the context of the vouchi.com voucher payment network.
 * However it does not support chat or presence via Jabber.
 * It also provides support for the Exchanges and Marketplaces tabs.
 *
 * @author Justin Turrell
 */
public final class WalletClient extends JFrame implements Runnable {
	// BEGIN data members
	/**
	 * name of directory in which log, preference, and token files are stored
	 */
	public static final String	M_VSCSubdir = "silentvault";
	
	/**
	 * default OFS hostname
	 */
	private static final String	M_DefaultOFShost = "ofs.voucher-safe.org";

	/**
	 * default OFS port
	 */
	private static final int	M_DefaultOFSport = 5233;

	/**
	 * default OFS XMPP domain (same as default host)
	 */
	private static final String	M_DefaultOFSdomain = "ofs.voucher-safe.org";

	/**
	 * version string displayed in Help/About
	 */
	private static final String	M_Version
										= " SilentVault Client rev. 0.7.1\n"
											+ " Built 22 March, 2017\n\n";

	/**
	 * path to help icon
	 */
	private static final String	M_HelpIconName = "images/help16x16Flat.png";

	/**
	 * path to info icon
	 */
	public static final	String	M_InfoIconName = "images/info16x16Flat.png";
	
	/**
	 * path to Voucher-Safe mini icon
	 */
	private static final	String	M_VSIconName = "images/VS_Logo.png";
	
	/**
	 * path to SV mini icon
	 */
	private static final String	M_SVIconName = "images/SV_Padlock.png";

	/**
	 * maximum number of configurable OFSen
	 */
	private static final int	M_MaxOFSConfig = 10;

	/**
	 * the actual help icon, inside a label
	 */
	public final JLabel		M_HelpIcon;

	/**
	 * the actual VS project icon, inside a label
	 */
	public final JLabel		M_VSIcon;

	/**
	 * the actual SV project icon, inside a label
	 */
	public final JLabel		M_SVIcon;

	/**
	 * default font to use on buttons
	 */
	public final Font			M_ButtonFont;

	/**
	 * the directory where logs, token cache, and preferences are stored
	 */
	private File				m_LogDir;

	/**
	 * default background to use on non-interactive (label) text fields
	 */
	public Color				m_LabelBackColor;

	/**
	 * the voucher system state-tracking object
	 */
	private VscState			m_StateData;

	/**
	 * the object for storing login data (secrets)
	 */
	private VsSecrets			m_LoginSecrets;

	/**
	 * the time at which this application last connected to the OFS gateway
	 */
	private Date				m_InitTimestamp;

	/**
	 * the IQ provider manager object (from Smack)
	 */
	private ProviderManager	m_ProviderManager;

	/**
	 * the object that listens for incoming OFS packets (replies)
	 */
	private OFSListener		m_OFSListener;

	/**
	 * the object that listens for incoming SVX packets (replies and broadcasts)
	 */
	private SVXListener		m_SVXListener;

	/**
	 * the object that listens for incoming SVM packets (replies and broadcasts)
	 */
	private SVMListener		m_SVMListener;

	/**
	 * the object that listens for incoming PKS packets (replies)
	 */
	private PKSListener		m_PKSListener;

	/**
	 * the object that listens for incoming DHT packets (replies)
	 */
	private DHTListener		m_DHTListener;

	/**
	 * the object that listens for incoming SDS packets (replies)
	 */
	private SDSListener		m_SDSListener;

	/**
	 * the object that listens for incoming website order messages
	 */
	private WebListener		m_WebListener;

	/**
	 * the filter which identifies website order messages
	 */
	private PacketFilter		m_WebMsgFilter;

	/**
	 * the filter which identifies SVM ticker and broadcast messages
	 */
	private PacketFilter		m_SVMMsgFilter;

	/**
	 * the filter which identifies SVX broadcast messages
	 */
	private PacketFilter		m_SVXMsgFilter;

	/**
	 * UI screen manager and main container for VS wallet tab
	 */
	private WTabManager		m_WalletTabManager;

	/**
	 * UI screen manager and main container for Exchanges tab
	 */
	private XTabManager		m_ExchangeTabManager;

	/**
	 * UI screen manager and main container for Marketplaces tab
	 */
	private MTabManager		m_MarketplaceTabManager;

	/**
	 * status/progress bar
	 */
	private JPanel				m_StatusBar;

	/**
	 * tabbed pane to contain the rest of the UI in 3 tabs
	 */
	private JTabbedPane		m_TabbedPane;

	/**
	 * popup frame for manual OFS connection data entry
	 */
	private JFrame				m_ManualEntry;

	/**
	 * entry for host or IP of manually specified OFS
	 */
	private JTextField		m_ManualHost;

	/**
	 * entry for port number of manually specified OFS
	 */
	private JTextField		m_ManualPort;

	/**
	 * entry for XMPP domain of manually specified OFS
	 */
	private JTextField		m_ManualXMPP;

	/**
	 * spent token list, to track tokens offered to SDS/DHT nodes but not yet
	 * cleared out of our folder
	 */
	private SpentTokenCache	m_SpentCache;

	/**
	 * the object which fetches and caches current asset prices
	 */
	private AssetPriceCache	m_PricingModule;

	/**
	 * app-specific user preferences
	 */
	private VSPreferences	m_UserPrefs;

	/**
	 * the timer which handles logging out on inactivity
	 */
	private java.util.Timer	m_InactivityTimer;

	/**
	 * the active connection to the OFS
	 */
	private XMPPConnection	m_OFSConnection;

	/**
	 * debug/error/warning logger
	 */
	private Log					m_Log;

	/**
	 * inner class to provide inactivity timer support
	 */
	protected class InactivityTimer extends TimerTask {
		public InactivityTimer() {
			super();
		}

		// method to extend the TimerTask class
		/**
		 * perform the action to be run by the inactivity timer
		 */
		public void run() {
			if (m_StateData == null) {
				Log.error("Inactivity timer has no state data available");
				return;
			}
			if (m_StateData == null || !m_StateData.isLoggedIn()) {
				// nothing to do
				return;
			}

			// get the timeout interval from user preferences
			int inactMin = m_UserPrefs.getInactivity();
			if (m_StateData.isExpired(inactMin * 60)) {
				// perform logout and return to login screen
				doLogout();

				/* Because we are now using PageKites to the OFS (and other
				 * servers), and because these grow stale a short time after
				 * :80 is pinged, and cause reconnect attempts to fail, we need
				 * to terminate the OFS connection when the wallet logs out.
				 */
				if (m_OFSConnection != null) {
					doDisconnect();
				}
				m_WalletTabManager.queueTransition(
													new Integer(WTabManager.M_LOGOUT_SAFE));
				m_WalletTabManager.makeTransition();
			}
		}
	}

	/**
	 * inner class to provide listener for window closure
	 */
	private class WinCloser extends WindowAdapter {
		public void windowClosing(WindowEvent we) {
			// log out, forcing token purge if needed
			if (m_StateData != null && m_StateData.isLoggedIn()) {
				doLogout();
			}

			// shut down app neatly
			shutdown();
			System.exit(0);
		}
	}

	/**
	 * inactivity timer support object
	 */
	private InactivityTimer m_TimerClass;

	/**
	 * inner class to provide OFS connection listener for status bar updates
	 */
	private class OFSConnListener implements ConnectionListener {
		public final Font		M_LabFont;
		private JLabel			m_ConLab;
		private WalletClient	m_Plugin;

		/**
		 * constructor
		 * @param plugin the application of which we're a part
		 */
		public OFSConnListener(WalletClient plugin) {
			m_Plugin = plugin;
			M_LabFont = new Font(Font.SANS_SERIF, Font.BOLD, 13);
			m_ConLab = new JLabel();
			m_ConLab.setFont(M_LabFont);
		}

		/**
		 * handle reconnection abort notification and close on fatal error
		 */
		public void connectionClosed() {
			/* NB: regular connect closed does not come here, because we remove
			 * the listener first.  Therefore this is unused except in cases of
			 * catastrophic errors where Smack closes our OFS port.  (Though one
			 * would think the latter cases would go to connectionClosedOnError().)
			 */
			Log.debug("OFS connection closed");
			connectionClosedOnError(null);
		}

		/**
		 * handle abnormal connection closed notification
		 * @param excep the exception which caused the closure, if any
		 */
		public void connectionClosedOnError(Exception excep) {
			m_ConLab.setText(" OFS connection lost - Not ready");
			m_StatusBar.removeAll();
			m_StatusBar.add(m_ConLab);
			m_StatusBar.add(Box.createHorizontalStrut(300));
			m_StatusBar.revalidate();
			m_StatusBar.repaint();
		}

		/**
		 * handle pending reconnection notification
		 * @param secs the number of seconds till reconnect attempt is made
		 */
		public void reconnectingIn(int secs) {
			if (secs <= 7) {
				m_ConLab.setText(" Reconnecting to OFS in " + secs + " seconds");
				m_StatusBar.removeAll();
				m_StatusBar.add(m_ConLab);
				m_StatusBar.add(Box.createHorizontalStrut(250));
				m_StatusBar.revalidate();
				m_StatusBar.repaint();
			}
		}

		/**
		 * handle successful reconnection notification
		 */
		public void reconnectionSuccessful() {
			setDefaultStatusBar();
			if (m_StateData != null && !m_StateData.isLoggedIn()) {
				m_WalletTabManager.queueTransition(
															new Integer(WTabManager.M_LOGIN1));
				SwingUtilities.invokeLater(m_Plugin);
			}
			// reset init timestamp
			m_InitTimestamp = new Date();
		}

		/**
		 * handle reconnection failure notification
		 * @param excep the exception which caused the failure
		 */
		public void reconnectionFailed(Exception excep) {
			m_ConLab.setText(" OFS reconnection failed - Not ready");
			m_StatusBar.removeAll();
			m_StatusBar.add(m_ConLab);
			m_StatusBar.add(Box.createHorizontalStrut(240));
			m_StatusBar.revalidate();
			m_StatusBar.repaint();

			// now force a logout if needed, and show disconnected screen
			if (m_StateData != null && m_StateData.isLoggedIn()) {
				doLogout();
			}
			m_WalletTabManager.queueTransition(
													new Integer(WTabManager.M_DISCONNECTED));
			SwingUtilities.invokeLater(m_Plugin);
		}
	}

	/**
	 * listener for connect events on the OFS port
	 */
	private OFSConnListener	m_OFSConnectListener;

	/**
	 * action for making a connection to the OFS
	 */
	private AbstractAction	m_ConnectAction;

	/**
	 * action for disconnecting from the OFS
	 */
	private AbstractAction	m_DisconnectAction;

	/**
	 * action for exiting
	 */
	private AbstractAction	m_ExitAction;

	/**
	 * action for going to preferences screen
	 */
	private AbstractAction	m_PreferencesAction;

	/**
	 * action for logout from a VS
	 */
	private AbstractAction	m_LogoutAction;

	/**
	 * action for JID display
	 */
	private AbstractAction	m_ShowJIDAction;

	/**
	 * action for JID get (to clipboard)
	 */
	private AbstractAction	m_GetJIDAction;

	/**
	 * action for closing a VS
	 */
	private AbstractAction	m_CloseSafeAction;

	/**
	 * action for displaying tokens
	 */
	private AbstractAction	m_ShowTokensAction;

	/**
	 * action for displaying asset prices
	 */
	private AbstractAction	m_ShowPricesAction;

	/**
	 * action for displaying tutorial help
	 */
	private AbstractAction	m_TutorialHelpAction;

	/**
	 * action for displaying About version string
	 */
	private AbstractAction	m_AboutAction;

	/**
	 * action to validate a voucher
	 */
	private AbstractAction	m_ValidateAction;

	/**
	 * action to split a voucher
	 */
	private AbstractAction	m_SplitAction;

	/**
	 * action to merge 2 or more vouchers
	 */
	private AbstractAction	m_MergeAction;

	/**
	 * action to deal with expired vouchers
	 */
	private AbstractAction	m_ExpiredAction;

	/**
	 * action to parse and attempt connection to a manually entered OFS gateway
	 */
	private AbstractAction	m_ManualConnectAction;

	/**
	 * inner class to describe an available OFS and its connect path
	 */
	protected final class OFSPathDescriptor {
		/**
		 * the XMPP domain of the OFS server
		 */
		public String					m_XMPPDomain;

		/**
		 * the path to the OFS, as a hostname and port
		 */
		public InetSocketAddress	m_AccessPath;

		/**
		 * whether the OFS's endpoint appeared valid
		 */
		public boolean					m_IsValid;

		/**
		 * constructor
		 * @param domain the XMPP domain
		 * @param host the hostname of the OFS's public endpoint
		 * @param port the port number @host accessing the OFS's c2s listen port
		 */
		public OFSPathDescriptor(String domain, String host, int port) {
			if (domain != null && !domain.isEmpty()) {
				m_XMPPDomain = new String(domain);
				try {
					m_AccessPath = new InetSocketAddress(host, port);
					m_IsValid = true;
				}
				catch (IllegalArgumentException iae) {
					Log.error("OFS host:port is not valid", iae);
				}
			}
		}
	}

	/**
	 * available hostnames for defined OFSen
	 */
	private ArrayList<OFSPathDescriptor>	m_AvailOFSHosts;

	/**
	 * the index of the selected OFS in m_AvailOFSHosts
	 */
	private int					m_OFSHostIndex;

	/**
	 * hostname of the OFS (currently selected from paths in m_AvailOFSHosts)
	 */
	private String				m_OFShost;

	/**
	 * port number of the OFS (currently selected from paths in m_AvailOFSHosts)
	 */
	private int					m_OFSport;

	/**
	 * XMPP domain name of the OFS (currently selected from paths in
	 * m_AvailOFSHosts)
	 */
	private String				m_OFSdomain;

	/**
	 * list of popup windows which must all be closed on VS logout
	 */
	private HashSet<JFrame>	m_PopupList;

	/**
	 * latest OFS connect attempt status
	 */
	private boolean			m_OFSConnectStat;

	/**
	 * whether we're running on a test network (i.e. @voucher-safe.org)
	 */
	private boolean			m_TestNet;

	/**
	 * the plugin specifying the SVM interface to the currently connected SVM
	 */
	private SVMPlugin			m_SVMPlugin;

	// END data members

	// BEGIN methods
	/**
	 * contructor
	 */
	public WalletClient() {
		super();
		setTitle("SilentVault");

		// init error logging
		String home = null;
		String os = null;
		try {
			os = System.getProperty("os.name");
			home = System.getProperty("user.home");
		}
		catch (SecurityException se) {
			System.err.println("Could not look up user home: " + se.getMessage());
			// try for defaults
			os = "Linux";
			home = "$HOME";
		}
		boolean onWin = os.startsWith("Win");
		String logDir = null;
		String altLogDir = null;
		if (onWin) {
			/* Check the profile first, then the properties default.  Note that
			 * both could fail, e.g. on Windows7 where the Desktop folder has
			 * been moved.
			 */
			logDir = System.getenv("USERPROFILE") + File.separator + M_VSCSubdir;
			altLogDir = home + File.separator + M_VSCSubdir;
		}
		else {
			logDir = home + File.separator + "." + M_VSCSubdir;
			altLogDir = logDir;
		}
		File logD = new File(logDir);
		if (!logD.exists()) {
			try {
				if (!logD.mkdirs()) {
					if (onWin) {
						// try alternate before failing
						logD = new File(altLogDir);
						if (!logD.exists()) {
							if (!logD.mkdirs()) {
								System.err.println("Could not create logfile directory "
														+ logDir + " or " + altLogDir);
							}
						}
					}
					else {
						System.err.println("Could not create logfile directory "
												+ logDir);
					}
				}
			}
			catch (SecurityException se) {
				System.err.println("Error creating log dir: " + se.getMessage());
			}
		}
		String logFile = logDir + File.separator + "errors.log";
		m_Log = new Log(logFile, this);
		m_Log.Init();
		m_LogDir = logD;

		// init login blob and network config data
		m_LoginSecrets = new VsSecrets();

		// define cache for current asset prices
		m_PricingModule = new AssetPriceCache(this);

		// obtain preferences
		m_UserPrefs = new VSPreferences(this);
		m_UserPrefs.load();

		// load standard icons
		ImageIcon helpIcon = null;
		ImageIcon projIcon = null;
		ImageIcon svIcon = null;
		URL imgURL = getClass().getResource(M_HelpIconName);
		if (imgURL == null) {
			Log.error("Could not find " + M_HelpIconName);
		}
		else {
			helpIcon = new ImageIcon(imgURL, "help icon");
		}
		M_HelpIcon = new JLabel(helpIcon);
		imgURL = getClass().getResource(M_VSIconName);
		if (imgURL == null) {
			Log.error("Could not find " + M_VSIconName);
		}
		else {
			projIcon = new ImageIcon(imgURL, "Voucher-Safe icon");
		}
		M_VSIcon = new JLabel(projIcon);
		imgURL = getClass().getResource(M_SVIconName);
		if (imgURL == null) {
			Log.error("Could not find " + M_SVIconName);
		}
		else {
			svIcon = new ImageIcon(imgURL, "SilentVault icon");
		}
		M_SVIcon = new JLabel(svIcon);

		m_ProviderManager = ProviderManager.getInstance();
		m_TimerClass = new InactivityTimer();
		m_WalletTabManager = new WTabManager(this);
		m_ExchangeTabManager = new XTabManager(this);
		// supply default SVM tab manager
		m_MarketplaceTabManager = new MTabManager(this);
		m_AvailOFSHosts = new ArrayList<OFSPathDescriptor>(M_MaxOFSConfig);
		m_OFShost = new String(M_DefaultOFShost);
		m_OFSport = M_DefaultOFSport;
		m_OFSdomain = new String(M_DefaultOFSdomain);
		M_ButtonFont = new Font(Font.SANS_SERIF, Font.BOLD, 12);
		m_LabelBackColor = new Color(214, 217, 223);
		m_OFSConnectListener = new OFSConnListener(this);
		m_PopupList = new HashSet<JFrame>();

		// provide standard border and label color
		Color stndGrey = Color.LIGHT_GRAY.darker();
		UIManager.put("standardGray", stndGrey);
		// set grey color for Nimbus progress bars also, by overriding orange
		UIManager.put("nimbusOrange", stndGrey);

		// create empty status bar panel
		m_StatusBar = new JPanel();
		m_StatusBar.setLayout(new BoxLayout(m_StatusBar, BoxLayout.X_AXIS));
		m_StatusBar.setPreferredSize(new Dimension(525, 20));
		m_StatusBar.setVisible(true);
		m_StatusBar.setBorder(BorderFactory.createLineBorder(stndGrey, 1));
		m_StatusBar.setBackground(Color.LIGHT_GRAY);

		// create tabbed pane
		m_TabbedPane = new JTabbedPane(JTabbedPane.BOTTOM);
	}

	/**
	 * main
	 * @param args array of arguments passed with invocation
	 */
	public static void main(String[] args) {
		WalletClient vscApp = new WalletClient();
		vscApp.initialize(args);
	}

   /**
	 * Called by main() to initialize the client.
	 * @param args arguments passed on the command line, if any
	 */
   public void initialize(String[] args) {
      //System.out.println("Welcome to the" + M_Version + ".");

		// init state data
		m_StateData = new VscState();

		// examine arguments to see if a properties file was specified
		String propsFile = "";
		for (int iii = 0; iii < args.length; iii++) {
			String arg = args[iii];
			// only argument we recognize is "-p <propertiesFile>"
			if (arg.equals("-p")) {
				// next arg will be properties file
				if (args.length > (iii+1)) {
					propsFile = args[iii+1];
					continue;
				}
			}
		}
		Properties propsConfig = new Properties();
		if (propsFile != null && !propsFile.isEmpty()) {
			FileInputStream fis = null;
			try {
				fis = new FileInputStream(propsFile);
				propsConfig.load(fis);
				fis.close();
			}
			catch (FileNotFoundException fnfe) {
				System.err.println("Could not find properties file, " + propsFile);
				System.exit(2);
			}
			catch (IOException ioe) {
				System.err.println("Error reading properties file, " + propsFile);
				System.exit(3);
			}
		}

		// initialze provider manager to know about our IQ provider classes
		OFSMessageProvider ofsProvider = new OFSMessageProvider();
		m_ProviderManager.addIQProvider(	"vsc_ofs",
													"jabber:iq:voucher-safe#vsc-ofs",
													ofsProvider	);
		PKSMessageProvider pksProvider = new PKSMessageProvider();
		m_ProviderManager.addIQProvider(	"sys_pks",
													"jabber:iq:voucher-safe#sys-pks",
													pksProvider	);
		DHTMessageProvider dhtProvider = new DHTMessageProvider();
		m_ProviderManager.addIQProvider(	"vsc_dht",
													"jabber:iq:voucher-safe#vsc-dht",
													dhtProvider	);
		SDSMessageProvider sdsProvider = new SDSMessageProvider();
		m_ProviderManager.addIQProvider(	"vsc_sds",
													"jabber:iq:voucher-safe#vsc-sds",
													sdsProvider	);
		SVGSVMMessageProvider svmProvider = new SVGSVMMessageProvider();
		m_ProviderManager.addIQProvider(	"vsc_ofs",
													"jabber:iq:silent-market#vsc-ofs",
													svmProvider);
		SVGSVXMessageProvider svxProvider = new SVGSVXMessageProvider();
		m_ProviderManager.addIQProvider(	"vsc_ofs",
													"jabber:iq:silent-vault#vsc-ofs",
													svxProvider);

		// initialize provider manager to know about our PacketExtension providers
		WebMessageProvider webProvider = new WebMessageProvider();
		m_ProviderManager.addExtensionProvider("web_vsc",
														"jabber:message:voucher-safe#web-vsc",
															webProvider );
		m_ProviderManager.addExtensionProvider("vsc_ofs",
													"jabber:message:silent-market#vsc-ofs",
															svmProvider );
		m_ProviderManager.addExtensionProvider("vsc_ofs",
														"jabber:message:silent-vault#vsc-ofs",
															svxProvider );

		// initialize listeners for IQ packets from partners
		m_OFSListener = new OFSListener(this);
		m_PKSListener = new PKSListener(this);
		m_DHTListener = new DHTListener(this);
		m_SDSListener = new SDSListener(this);
		m_SVXListener = new SVXListener(this);
		// supply default SVM listener
		m_SVMListener = new SVMListener(this);
		// supply default SVM plugin configuration
		m_SVMPlugin = new SVMPlugin(this);
		m_SVMPlugin.setSVMListener(m_SVMListener);
		m_SVMPlugin.setClientBlock("org.silentvault.client.SVMClientBlock");
		m_SVMPlugin.setReplyBlock("org.silentvault.client.SVMBlock");
		m_SVMPlugin.setTabManager(m_MarketplaceTabManager);

		// init listener for website packets
		m_WebListener = new WebListener();

		// set up filter to listen for Message packets from websites
		m_WebMsgFilter = new PacketFilter() {
			public boolean accept(Packet packet) {
				if (packet instanceof Message) {
					Message msg = (Message) packet;
					if (msg.getType() == Message.Type.normal) {
						// see if we have an extension present
						if (msg.getExtension("web_vsc",
													"jabber:message:voucher-safe#web-vsc")
							!= null)
						{
							return true;
						}
					}
				}
				return false;
			}
		};

		// set up filter to listen for broadcast and ticker packets from an SVM
		m_SVMMsgFilter = new PacketFilter() {
			public boolean accept(Packet packet) {
				if (packet instanceof Message) {
					Message msg = (Message) packet;
					if (msg.getType() == Message.Type.normal) {
						// see if we have the SVM extension present
						if (msg.getExtension("vsc_ofs",
													"jabber:message:silent-market#vsc-ofs")
							!= null)
						{
							return true;
						}
					}
				}
				return false;
			}
		};

		// set up filter to listen for broadcast packets from an SVX
		m_SVXMsgFilter = new PacketFilter() {
			public boolean accept(Packet packet) {
				if (packet instanceof Message) {
					Message msg = (Message) packet;
					if (msg.getType() == Message.Type.normal) {
						// see if we have the SVX extension present
						if (msg.getExtension("vsc_ofs",
													"jabber:message:silent-vault#vsc-ofs")
							!= null)
						{
							return true;
						}
					}
				}
				return false;
			}
		};

		// define spent tokens cache
		m_SpentCache = new SpentTokenCache(this);

		// examine properties to set OFS coordinates and error recipient
		Properties props = null;
		try {
			if (propsConfig.size() > 0) {
				// use properties file specified on command line
				props = propsConfig;
			}
			else {
				// use properties passed from .jnlp
				props = System.getProperties();
			}

			// get test network property
			String testNet = props.getProperty("TestNet");
			if (testNet != null) {
				m_TestNet = Boolean.parseBoolean(testNet);
			}

			// get CC recipient for client errors logged, if any
			String edest = props.getProperty("ErrorJID");
			if (edest != null) {
				m_Log.setErrRecip(edest);
			}

			// OFS coordinates, host and port number
			String ppp = props.getProperty("OFShost");
			String defOFShost = null;
			if (ppp != null) {
				defOFShost = new String(ppp);
			}
			else {
				Log.error("Missing OFS hostname property, using default");
			}
			int defOFSport = 0;
			ppp = props.getProperty("OFSport");
			if (ppp != null) {
				Integer pnum = null;
				try {
					pnum = Integer.parseInt(ppp);
					defOFSport = pnum.intValue();
				}
				catch (NumberFormatException nfe) {
					Log.error("Invalid OFS port " + ppp, nfe);
				}
			}
			else {
				Log.error("Missing OFS port property, using default");
			}
			String xmpp = props.getProperty("OFSxmpp");
			String defOFSXMPP = defOFShost;
			if (xmpp != null && !xmpp.isEmpty()) {
				defOFSXMPP = new String(xmpp);
			}
			else {
				Log.error("Missing OFS domain property, using default (== host)");
			}

			// put the default into the available OFS list
			OFSPathDescriptor defOFS
				= new OFSPathDescriptor(defOFSXMPP, defOFShost, defOFSport);
			if (defOFS.m_IsValid) {
				m_AvailOFSHosts.add(defOFS);
			}

			/* Now, if any additional OFSen are specified (as OFShost01, OFShost02,
			 * plus matching OFSport01, OFSport02, etc.), add them to the list.
			 */
			for (int ofs = 1; ofs < M_MaxOFSConfig; ofs++) { 
				String hostProp = "OFShost0" + ofs;
				String portProp = "OFSport0" + ofs;
				String xmppProp = "OFSxmpp0" + ofs;
				String ofsHost = props.getProperty(hostProp);
				String ofsPort = props.getProperty(portProp);
				String ofsXmpp = props.getProperty(xmppProp);
				if (ofsXmpp != null && ofsHost != null && ofsPort != null) {
					// try to make an address out of this
					OFSPathDescriptor ofsAddr = null;
					Integer pNum = null;
					try {
						pNum = Integer.parseInt(ofsPort);
						ofsAddr
							= new OFSPathDescriptor(ofsXmpp, ofsHost, pNum.intValue());
						if (ofsAddr.m_IsValid) {
							m_AvailOFSHosts.add(ofsAddr);
						}
					}
					catch (NumberFormatException nfe) {
						Log.error("Invalid port, " + ofsPort, nfe);
					}
					catch (IllegalArgumentException iae) {
						Log.error("Defined OFS is not valid: " + ofsXmpp + ":"
									+ ofsHost + ":" + ofsPort, iae);
					}
				}
			}

			// supply compiled-in default if we have none defined
			if (m_AvailOFSHosts.isEmpty()) {
				defOFS = new OFSPathDescriptor(m_OFSdomain, m_OFShost, m_OFSport);
				m_AvailOFSHosts.add(defOFS);
			}

			// take first OFS added to list as default
			m_OFSHostIndex = 0;
			defOFS = m_AvailOFSHosts.get(m_OFSHostIndex);
			InetSocketAddress defaultOFS = defOFS.m_AccessPath;
			m_OFShost = defaultOFS.getHostString();
			m_OFSport = defaultOFS.getPort();
			m_OFSdomain = defOFS.m_XMPPDomain;
		}
		catch (SecurityException se) {
			Log.error("Cannot fetch OFS domain, hostname and port from system "
						+ "properties", se);
		}

		// if requested in preferences, init connection to the OFS
		if (m_UserPrefs.getConnectAtStart()) {
			m_OFSConnection = openOFS(m_OFSdomain, m_OFShost, m_OFSport);
			if (m_OFSConnection == null) {
				Log.error("Unable to open connection to OFS");
				setDefaultStatusBar();
			}
			else {
				m_OFSConnection.addConnectionListener(m_OFSConnectListener);
				m_OFSConnection.addPacketListener(m_WebListener, m_WebMsgFilter);
				setSVMListener(null);
				m_OFSConnection.addPacketListener(m_SVXListener, m_SVXMsgFilter);

				// do disco to find the SVXen and SVMs
				if (!doDisco()) {
					Log.error("Error doing SVX/SVM discovery");
				}

				/* NB: we do not invoke run() here directly, because we don't want
				 * 	 to manipulate the stat bar until after initUI() has returned.
				 */
				m_PricingModule.initPriceFetching();
			}
		}

		// activate inactivity timer, checking once a minute
		m_InactivityTimer = new java.util.Timer("inactivity");
		m_InactivityTimer.schedule(m_TimerClass, 60000L, 60000L);

		// add UI elements (WTabManager), via a thread scheduled to run later
		SwingUtilities.invokeLater(new Runnable() {
			public void run() {
		   	initUI();

				// now that the UI is set up, perform first prices fetch
				m_PricingModule.run();
			}
		});

		//Log.debug("Silent-Vault Client Initialized");
   }

   /**
	 * Called when shutting down to allow for persistence of
	 * information or releasing of resources.
 	 */
   public void shutdown() {
		// nix inactivity timer
		if (m_InactivityTimer != null) {
			m_InactivityTimer.cancel();
			m_InactivityTimer = null;
		}
		m_StateData = null;

		// abolish specific namespace IQ providers
		m_ProviderManager.removeIQProvider("vsc_ofs",
														"jabber:iq:voucher-safe#vsc-ofs");
		m_ProviderManager.removeIQProvider("sys_pks",
														"jabber:iq:voucher-safe#sys-pks");
		m_ProviderManager.removeIQProvider("vsc_dht",
														"jabber:iq:voucher-safe#vsc-dht");
		m_ProviderManager.removeIQProvider("vsc_sds",
														"jabber:iq:voucher-safe#vsc-sds");
		m_ProviderManager.removeIQProvider("vsc_ofs",
														"jabber:iq:silent-market#vsc-ofs");
		m_ProviderManager.removeIQProvider("vsc_ofs",
														"jabber:iq:silent-vault#vsc-ofs");
		// abolish extension providers
		m_ProviderManager.removeExtensionProvider("web_vsc",
													"jabber:message:voucher-safe#web-vsc");
		m_ProviderManager.removeExtensionProvider("vsc_ofs",
													"jabber:message:silent-market#vsc-ofs");
		m_ProviderManager.removeExtensionProvider("vsc_ofs",
													"jabber:message:silent-vault#vsc-ofs");

		// stop listening for OFS reply packets
		if (m_OFSConnection != null) {
			doDisconnect();
		}
		m_OFSListener = null;
		m_AvailOFSHosts.clear();
		if (m_PKSListener != null) {
			m_PKSListener.closePKSConnection();
		}
		m_PKSListener = null;
		if (m_DHTListener != null) {
			m_DHTListener.closeDHTConnection();
		}
		m_DHTListener = null;
		if (m_SDSListener != null) {
			m_SDSListener.closeSDSConnection();
		}
		m_SDSListener = null;
		m_WebListener = null;
		m_WebMsgFilter = null;
		m_SVMMsgFilter = null;
		m_SVXMsgFilter = null;
		m_SVMListener = null;
		m_SVXListener = null;
		m_ProviderManager = null;

		// get rid of any extant UI elements
		m_WalletTabManager.deallocatePanes();
		m_WalletTabManager = null;
		m_ExchangeTabManager.deallocatePanes();
		m_ExchangeTabManager = null;
		m_MarketplaceTabManager.deallocatePanes();
		m_MarketplaceTabManager = null;

		// sync spent token cache and then clear secrets
		if (m_SpentCache != null) {
			m_SpentCache.syncSpentTokens();
		}
		resetSecrets();

		//Log.debug("Silent-Vault Client De-initialized");
		m_Log.deInit();
   }

	/**
	 * reset the secrets object
	 */
	public void resetSecrets() { m_LoginSecrets = new VsSecrets(); }

	/**
	 * Initiates tab manager object and tells it to start with a login screen.
	 * Also defines main menu.
	 */
	private void initUI() {
		setLayout(new GridLayout());
		addWindowListener(new WinCloser());
		setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
		final Dimension appSz = new Dimension(540, 540);
		setPreferredSize(appSz);

		// position window in center of screen
		GraphicsEnvironment g = GraphicsEnvironment.getLocalGraphicsEnvironment();
		Point center = g.getCenterPoint();
		double centerX = center.getX() - (appSz.getWidth() / 2.0);
		double centerY = center.getY() - (appSz.getHeight() / 2.0);
		Point topLeft = new Point((int) centerX, (int) centerY);
		setLocation(topLeft);

		// specify preference for proportional fonts, to help make layouts
		// consistent across various platforms
		g.preferProportionalFonts();

		// create a pane to hold the tabbed pane plus the status bar
		JPanel contents = new JPanel();
		contents.setLayout(new BoxLayout(contents, BoxLayout.Y_AXIS));
		// create tabbed pane with our 3 tab managers as the Components
		//m_TabbedPane = new JTabbedPane(JTabbedPane.BOTTOM);
		m_TabbedPane.addTab("Wallet ", M_VSIcon.getIcon(), m_WalletTabManager,
									"Perform wallet operations");
		m_TabbedPane.setMnemonicAt(0, KeyEvent.VK_1);
		m_TabbedPane.addTab("Exchanges ", M_SVIcon.getIcon(),
									m_ExchangeTabManager,
									"SilentVault Exchange operations");
		m_TabbedPane.setMnemonicAt(1, KeyEvent.VK_2);
		m_TabbedPane.addTab("Marketplaces ", M_SVIcon.getIcon(),
									m_MarketplaceTabManager,
									"SilentVault Market operations");
		m_TabbedPane.setMnemonicAt(2, KeyEvent.VK_3);
		/* NB: setting no dragNdrop target mitigates a Sun Java bug which
		 * 	 causes this exception to be thrown (on Linux only):
		 * 	 sun.awt.X11.XException: Cannot write XdndAware property
		 */
		contents.setDropTarget(null);
		m_WalletTabManager.setDropTarget(null);
		m_ExchangeTabManager.setDropTarget(null);
		m_MarketplaceTabManager.setDropTarget(null);
		m_TabbedPane.setDropTarget(null);
		m_StatusBar.setDropTarget(null);
		setDropTarget(null);
		contents.add(m_TabbedPane);
		contents.add(m_StatusBar);
		contents.setBorder(null);
		add(contents);

		// establish UI entries for manual OFS gateway entry
		final WalletClient SC = this;
		m_ManualHost = new JTextField(32);
		m_ManualHost.setDropTarget(null);
		m_ManualHost.setHorizontalAlignment(JTextField.LEFT);
		m_ManualHost.setToolTipText("Enter hostname or dotted quad IP address");
		m_ManualPort = new JTextField(6);
		m_ManualPort.setDropTarget(null);
		m_ManualPort.setHorizontalAlignment(JTextField.LEFT);
		m_ManualPort.setToolTipText("Enter integer port number at host");
		m_ManualXMPP = new JTextField(24);
		m_ManualXMPP.setDropTarget(null);
		m_ManualXMPP.setHorizontalAlignment(JTextField.LEFT);
		m_ManualXMPP.setToolTipText("Enter XMPP domain (defaults to host if "
											+ "left blank)");

		// main menu bar, with layout
		final JMenuBar menuBar = new JMenuBar();
		menuBar.setLayout(new BoxLayout(menuBar, BoxLayout.X_AXIS));

		// define File sub menu
		JMenu fileMenu = new JMenu("File");
		fileMenu.setMnemonic(KeyEvent.VK_F);
		
		// define action and menu item to switch to preferences screen
		m_PreferencesAction = new AbstractAction("Preferences") {
			public void actionPerformed(ActionEvent actionEvent) {
				m_WalletTabManager.queueTransition(
													new Integer(WTabManager.M_PREFERENCES));
				m_WalletTabManager.makeTransition();
			}
		};
		JMenuItem prefItem = new JMenuItem(m_PreferencesAction);
		prefItem.setMnemonic(KeyEvent.VK_P);
		prefItem.setIcon(SparkRes.getImageIcon(SparkRes.PREFERENCES_IMAGE));
		fileMenu.add(prefItem);

		// define action and menu item to exit application
		m_ExitAction = new AbstractAction("Exit") {
			public void actionPerformed(ActionEvent actionEvent) {
				// perform logout if needed
				if (m_StateData != null && m_StateData.isLoggedIn()) {
					doLogout();
				}
				
				// shut down app neatly
				shutdown();
				System.exit(0);
			}
		};
		JMenuItem exitItem = new JMenuItem(m_ExitAction);
		exitItem.setMnemonic(KeyEvent.VK_E);
		exitItem.setIcon(SparkRes.getImageIcon(SparkRes.SMALL_DELETE));
		fileMenu.add(exitItem);
		menuBar.add(fileMenu);
		menuBar.add(Box.createHorizontalStrut(10));

		// define action to parse and attempt to connect to manually entered OFS
		m_ManualConnectAction = new AbstractAction("Manual Connect") {
			public void actionPerformed(ActionEvent actionEvent) {
				// sanity check entries
				final String connErr = "cannot connect to manual coordinates";
				String hTxt = m_ManualHost.getText();
				if (hTxt.indexOf('.') == -1) {
					m_WalletTabManager.showError(connErr,
															"host or IP missing or invalid");
					return;
				}
				String dTxt = m_ManualXMPP.getText();
				if (dTxt.isEmpty()) {
					// default to host
					dTxt = hTxt;
				}
				if (dTxt.indexOf('.') == -1) {
					m_WalletTabManager.showError(connErr, "XMPP domain is invalid");
					return;
				}
				String pTxt = m_ManualPort.getText();
				Integer port = null;
				try {
					port = Integer.parseInt(pTxt);
				}
				catch (NumberFormatException nfe) {
					m_WalletTabManager.showError(connErr,
															"port number is not an integer");
					return;
				}
				if (port.intValue() <= 0) {
					m_WalletTabManager.showError(connErr,
														"port number must be greater than 0");
					return;
				}

				// try to make a path descriptor out of the entered data
				OFSPathDescriptor manualPath
					= new OFSPathDescriptor(dTxt, hTxt, port.intValue());
				if (!manualPath.m_IsValid) {
					m_WalletTabManager.showError(connErr,
														"manual coordinates are not valid");
					return;
				}

				if (m_OFSConnection != null && m_OFSConnection.isConnected()) {
					if (m_UserPrefs.getConnectAtStart()) {
						if (m_StateData != null && m_StateData.isLoggedIn()) {
							doLogout();
						}
						// disconnect them first
						doDisconnect();
					}
					else {
						// don't allow this if presently connected
						m_WalletTabManager.showError(connErr,
																"you are already connected",
																"please disconnect first");
						return;
					}
				}

				// attempt connection
				XMPPConnection manualConn = openOFS(manualPath.m_XMPPDomain,
													manualPath.m_AccessPath.getHostString(),
													manualPath.m_AccessPath.getPort());
				if (manualConn == null) {
					m_WalletTabManager.showError(connErr, "connection failed",
															"please check coordinates");
					m_WalletTabManager.queueTransition(
													new Integer(WTabManager.M_DISCONNECTED));
					m_WalletTabManager.makeTransition();
					setDefaultStatusBar();
				}
				else {
					m_OFSConnection = manualConn;
					m_OFShost = manualPath.m_AccessPath.getHostString();
					m_OFSport = manualPath.m_AccessPath.getPort();
					m_OFSdomain = manualPath.m_XMPPDomain;
					m_OFSConnection.addConnectionListener(m_OFSConnectListener);
					m_OFSConnection.addPacketListener(m_WebListener, m_WebMsgFilter);
					setSVMListener(null);
					m_OFSConnection.addPacketListener(m_SVXListener, m_SVXMsgFilter);
					deregisterPopup(m_ManualEntry);
					m_ManualEntry.dispose();

					// do disco to find the SVXen and SVMs
					if (!doDisco()) {
						Log.error("Error doing SVX/SVM discovery");
					}
					if (m_PricingModule == null) {
						m_PricingModule = new AssetPriceCache(SC);
					}
					m_PricingModule.run();
					m_PricingModule.initPriceFetching();
					m_WalletTabManager.queueTransition(
															new Integer(WTabManager.M_LOGIN1));
					m_WalletTabManager.makeTransition();
				}
			}
		};

		// connection menu
		JMenu connMenu = new JMenu("Connection ");
		connMenu.setMnemonic(KeyEvent.VK_C);

		// optional sub-menu to specify additional OFSen for selection
		JMenu ofsSelectMenu = new JMenu("Select OFS Gateway ");
		ofsSelectMenu.setIcon(SparkRes.getImageIcon(SparkRes.ADDRESS_BOOK_16x16));

		// build button menu items and define actions for each available OFS
		ButtonGroup bGroup = new ButtonGroup();
		for (int ofsIdx = 0; ofsIdx < m_AvailOFSHosts.size(); ofsIdx++) {
			OFSPathDescriptor ofsAddr = m_AvailOFSHosts.get(ofsIdx);
			AbstractAction selectAction = new AbstractAction(
												ofsAddr.m_AccessPath.getHostString()
												+ ":" + ofsAddr.m_AccessPath.getPort())
			{
				public void actionPerformed(ActionEvent actionEvent) {
					Object idxObj = this.getValue("index");
					if (idxObj != null && (idxObj instanceof Integer)) {
						Integer oIdx = (Integer) idxObj;
						int idx = oIdx.intValue();
						if (idx >= 0 && idx < m_AvailOFSHosts.size()) {
							m_OFSHostIndex = oIdx.intValue();
						}
					}
					// now perform the connection if the modal setting indicates
					if (m_UserPrefs.getConnectAtStart()) {
						if (m_OFSConnection != null
							 && m_OFSConnection.isConnected())
						{
							if (m_StateData != null && m_StateData.isLoggedIn()) {
								doLogout();
							}
							// disconnect them first
							doDisconnect();
						}
						// trigger regular connect action to use m_OFSHostIndex
						ActionEvent trigger = new ActionEvent(this, 0, "AUTOCONNECT");
						m_ConnectAction.actionPerformed(trigger);
					}
				}
			};
			selectAction.putValue("index", new Integer(ofsIdx));
			JRadioButtonMenuItem ofsButton
				= new JRadioButtonMenuItem(selectAction);
			bGroup.add(ofsButton);
			if (ofsIdx == m_OFSHostIndex) {
				ofsButton.setSelected(true);
			}
			else {
				ofsButton.setSelected(false);
			}
			ofsSelectMenu.add(ofsButton);
		}

		// add an item for manual entry of OFS coordinates
		AbstractAction manualEntry = new AbstractAction("Manual Entry") {
			public void actionPerformed(ActionEvent actionEvent) {
				// create a frame and put the connect dialog inside it
				if (m_ManualEntry != null) {
					// get rid of any old one first
					deregisterPopup(m_ManualEntry);
					m_ManualEntry.dispose();
				}
				m_ManualEntry = new JFrame("Manual Entry of Network Gateway");
				m_ManualEntry.setDefaultCloseOperation(
														WindowConstants.DISPOSE_ON_CLOSE);
				m_ManualEntry.setSize(new Dimension(350, 250));
				m_ManualEntry.setLocationRelativeTo(SC);
				registerPopup(m_ManualEntry);
				JPanel ofsPane = new JPanel();
				ofsPane.setLayout(new BoxLayout(ofsPane, BoxLayout.Y_AXIS));
				ofsPane.add(Box.createVerticalStrut(20));
				final Color sGrey = UIManager.getColor("standardGray");
				final JLabel hostLab = new JLabel("Gateway host or IP:");
				final Dimension lineSz = new Dimension(300, 25);

				// specify host or IP
				JPanel hostPane = new JPanel();
				hostPane.setLayout(new BoxLayout(hostPane, BoxLayout.X_AXIS));
				hostPane.setPreferredSize(lineSz);
				hostPane.add(Box.createHorizontalStrut(10));
				Font labFont = new Font(Font.SANS_SERIF, Font.PLAIN, 12);
				hostLab.setFont(labFont);
				hostLab.setForeground(sGrey);
				hostLab.setHorizontalAlignment(SwingConstants.LEFT);
				hostPane.add(hostLab);
				hostPane.add(Box.createHorizontalStrut(30));
				hostPane.add(m_ManualHost);
				hostPane.add(Box.createHorizontalStrut(10));
				ofsPane.add(hostPane);
				ofsPane.add(Box.createVerticalStrut(20));

				// specify port number
				JPanel portPane = new JPanel();
				portPane.setLayout(new BoxLayout(portPane, BoxLayout.X_AXIS));
				portPane.setPreferredSize(lineSz);
				portPane.add(Box.createHorizontalStrut(10));
				JLabel portLab = new JLabel("Port number:");
				portLab.setFont(labFont);
				portLab.setForeground(sGrey);
				portLab.setHorizontalAlignment(SwingConstants.LEFT);
				portPane.add(portLab);
				portPane.add(Box.createHorizontalStrut(65));
				portPane.add(m_ManualPort);
				portPane.add(Box.createHorizontalStrut(10));
				ofsPane.add(portPane);
				ofsPane.add(Box.createVerticalStrut(20));

				// specify XMPP of host
				JPanel xmppPane = new JPanel();
				xmppPane.setLayout(new BoxLayout(xmppPane, BoxLayout.X_AXIS));
				xmppPane.setPreferredSize(lineSz);
				xmppPane.add(Box.createHorizontalStrut(10));
				JLabel xmppLab = new JLabel("XMPP domain:");
				xmppLab.setFont(labFont);
				xmppLab.setForeground(sGrey);
				xmppLab.setHorizontalAlignment(SwingConstants.LEFT);
				xmppPane.add(xmppLab);
				xmppPane.add(Box.createHorizontalStrut(60));
				xmppPane.add(m_ManualXMPP);
				xmppPane.add(Box.createHorizontalStrut(10));
				ofsPane.add(xmppPane);
				ofsPane.add(Box.createVerticalStrut(20));

				// supply enter and cancel buttons
				JPanel buttonPane = new JPanel();
				buttonPane.setLayout(new BoxLayout(buttonPane, BoxLayout.X_AXIS));
				buttonPane.add(Box.createHorizontalStrut(20));
				AbstractAction cancelAction = new AbstractAction("Cancel") {
					public void actionPerformed(ActionEvent ae) {
						deregisterPopup(m_ManualEntry);
						m_ManualEntry.dispose();
					}
				};
				JButton cancel = new JButton(cancelAction);
				cancel.setFont(M_ButtonFont);
				cancel.setToolTipText("Abort manual entry");
				buttonPane.add(cancel);
				buttonPane.add(Box.createHorizontalStrut(50));
				JButton connect = new JButton(m_ManualConnectAction);
				connect.setToolTipText("Click here to connect to your manually "
												+ "specified network gateway");
				connect.setFont(M_ButtonFont);
				buttonPane.add(connect);
				ofsPane.add(buttonPane);
				ofsPane.add(Box.createVerticalStrut(20));
				m_ManualEntry.add(ofsPane);
				m_ManualEntry.setVisible(true);
			}
		};
		JRadioButtonMenuItem manualButton = new JRadioButtonMenuItem(manualEntry);
		manualButton.setSelected(false);
		bGroup.add(manualButton);
		ofsSelectMenu.add(manualButton);
		connMenu.add(ofsSelectMenu);

		// define action to connect to the OFS
		m_ConnectAction = new AbstractAction("Connect") {
			public void actionPerformed(ActionEvent actionEvent) {
				if (m_OFSConnection != null && m_OFSConnection.isConnected()) {
					m_WalletTabManager.showInfo("You are already connected");
					return;
				}
				OFSPathDescriptor ofsAddr = m_AvailOFSHosts.get(m_OFSHostIndex);
				m_OFShost = ofsAddr.m_AccessPath.getHostString();
				m_OFSport = ofsAddr.m_AccessPath.getPort();
				m_OFSdomain = ofsAddr.m_XMPPDomain;
				m_OFSConnection = openOFS(m_OFSdomain, m_OFShost, m_OFSport);
				if (m_OFSConnection == null) {
					m_WalletTabManager.queueTransition(
													new Integer(WTabManager.M_DISCONNECTED));
					m_WalletTabManager.makeTransition();
					setDefaultStatusBar();
				}
				else {
					m_OFSConnection.addConnectionListener(m_OFSConnectListener);
					m_OFSConnection.addPacketListener(m_WebListener, m_WebMsgFilter);
					setSVMListener(null);
					m_OFSConnection.addPacketListener(m_SVXListener, m_SVXMsgFilter);

					// do disco to find the SVXen and SVMs
					if (!doDisco()) {
						Log.error("Error doing SVX/SVM discovery");
					}
					if (m_PricingModule == null) {
						m_PricingModule = new AssetPriceCache(SC);
					}
					m_PricingModule.run();
					m_PricingModule.initPriceFetching();
					m_WalletTabManager.queueTransition(
															new Integer(WTabManager.M_LOGIN1));
					m_WalletTabManager.makeTransition();
				}
			}
		};
		JMenuItem connectItem = new JMenuItem(m_ConnectAction);
		connectItem.setMnemonic(KeyEvent.VK_C);
		connectItem.setIcon(SparkRes.getImageIcon(SparkRes.DOOR_IMAGE));
		connMenu.add(connectItem);

		// define action to disconnect from the OFS
		m_DisconnectAction = new AbstractAction("Disconnect") {
			public void actionPerformed(ActionEvent actionEvent) {
				if (m_OFSConnection != null) {
					// perform logout if needed
					if (m_StateData != null && m_StateData.isLoggedIn()) {
						doLogout();
					}
					doDisconnect();
					m_WalletTabManager.queueTransition(
													new Integer(WTabManager.M_DISCONNECTED));
					m_WalletTabManager.makeTransition();
				}
				else {
					m_WalletTabManager.showInfo("You are already disconnected");
				}
			}
		};
		JMenuItem disconnItem = new JMenuItem(m_DisconnectAction);
		disconnItem.setMnemonic(KeyEvent.VK_D);
		disconnItem.setIcon(SparkRes.getImageIcon(SparkRes.CLOSE_IMAGE));
		connMenu.add(disconnItem);
		menuBar.add(connMenu);
		menuBar.add(Box.createHorizontalStrut(10));

		// advanced menu
		JMenu advMenu = new JMenu("Advanced ");
		advMenu.setMnemonic(KeyEvent.VK_A);

		// define action to show JID
		m_ShowJIDAction = new AbstractAction("Show JID") {
			public void actionPerformed(ActionEvent actionEvent) {
				if (m_OFSConnection != null && m_OFSConnection.isConnected()) {
					// create a popup window which displays the user's JID
					JFrame jPop = new JFrame("Your Jabber ID");
					jPop.setDefaultCloseOperation(WindowConstants.HIDE_ON_CLOSE);
					JTextArea jid = new JTextArea(9, 30);
					jid.setEditable(false);
					jid.setLineWrap(true);
					// use same font as HomePane.M_DataFont
					jid.setFont(new Font(Font.SANS_SERIF, Font.PLAIN, 12));
					jid.setWrapStyleWord(true);
					StringBuilder jidTxt = new StringBuilder(128);
					jidTxt.append("Your current Jabber ID (JID) is:\n\n");
					String userJID = m_OFSConnection.getUser();
					int slash = userJID.lastIndexOf('/');
					jidTxt.append(userJID.substring(0, slash) + "\n\n");
					jidTxt.append("You may be asked for this to make a payment on ");
					jidTxt.append("a website which accepts vouchers.\n");
					jid.setText(jidTxt.toString());
					jPop.add(jid);
					jPop.setSize(new Dimension(230, 150));
					jPop.setLocationRelativeTo(menuBar);
					jPop.setVisible(true);
					registerPopup(jPop);
				}
				else {
					m_WalletTabManager.showInfo("You are not presently connected");
				}
			}
		};
		JMenuItem jidItem = new JMenuItem(m_ShowJIDAction);
		jidItem.setMnemonic(KeyEvent.VK_J);
		jidItem.setIcon(SparkRes.getImageIcon(SparkRes.SMALL_USER1_INFORMATION));
		jidItem.setToolTipText("Display your anonymous Jabber connect ID");
		advMenu.add(jidItem);

		// define action to copy JID to clipboard
		m_GetJIDAction = new AbstractAction("Get JID") {
			public void actionPerformed(ActionEvent actionEvent) {
				if (m_OFSConnection != null && m_OFSConnection.isConnected()) {
					String userJID = m_OFSConnection.getUser();
					int slash = userJID.lastIndexOf('/');
					StringSelection ss
						= new StringSelection(userJID.substring(0, slash));
					Toolkit tk = Toolkit.getDefaultToolkit();
					Clipboard sysClip = tk.getSystemClipboard();
					sysClip.setContents(ss, ss);
					m_WalletTabManager.showInfo("Your JID has been copied to the "
														+ "clipboard.\nYou may use this to "
														+ "make a payment on a website which "
														+ "accepts vouchers.");
				}
				else {
					m_WalletTabManager.showInfo("You are not presently connected");
				}
			}
		};
		JMenuItem jidItem2 = new JMenuItem(m_GetJIDAction);
		jidItem2.setMnemonic(KeyEvent.VK_G);
		jidItem2.setIcon(SparkRes.getImageIcon(SparkRes.NOTEBOOK_IMAGE));
		jidItem2.setToolTipText("Copy your anonymous Jabber connect ID to the "
										+ "clipboard");
		advMenu.add(jidItem2);

		// define action to show prices
		m_ShowPricesAction = new AbstractAction("Show Prices") {
			public void actionPerformed(ActionEvent actionEvent) {
				if (m_StateData != null && m_StateData.isLoggedIn()) {
					// go to prices screen
					m_WalletTabManager.queueTransition(
															new Integer(WTabManager.M_PRICES));
					m_WalletTabManager.makeTransition();
				}
				else {
					m_WalletTabManager.showError("cannot display prices",
														  "not logged in",
														"please log into your wallet first");
				}
			}
		};
		JMenuItem pricesItem = new JMenuItem(m_ShowPricesAction);
		pricesItem.setMnemonic(KeyEvent.VK_P);
		pricesItem.setToolTipText("Display current gold prices");
		pricesItem.setIcon(SparkRes.getImageIcon(SparkRes.DOWN_ARROW_IMAGE));
		advMenu.add(pricesItem);

		// define action to show tokens
		m_ShowTokensAction = new AbstractAction("Show Tokens") {
			public void actionPerformed(ActionEvent actionEvent) {
				if (m_StateData != null && m_StateData.isLoggedIn()) {
					// show token table in a popup
					JFrame tPop = new JFrame("Snapshot of Tokens in "
													+ m_LoginSecrets.getVSnumber());
					tPop.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
					JTable tokTab = m_WalletTabManager.getHomePane().getTokenTable();
					JScrollPane js = new JScrollPane(tokTab);
					tPop.setMinimumSize(new Dimension(455, 250));
					tPop.setPreferredSize(new Dimension(610, 400));
					tPop.add(js);
					tPop.setLocationRelativeTo(m_WalletTabManager);
					tPop.setVisible(true);
					registerPopup(tPop);
				}
				else {
					m_WalletTabManager.showInfo("You are not logged in to a wallet");
				}
			}
		};
		JMenuItem tokensItem = new JMenuItem(m_ShowTokensAction);
		tokensItem.setMnemonic(KeyEvent.VK_T);
		tokensItem.setToolTipText("Display the actual tokens in your safe");
		tokensItem.setIcon(SparkRes.getImageIcon(SparkRes.SMALL_DOCUMENT_VIEW));
		advMenu.add(tokensItem);

		// define action to validate a voucher
		m_ValidateAction = new AbstractAction("Validate") {
			public void actionPerformed(ActionEvent actionEvent) {
				// get rid of the voucher list popup if one is displayed
				MenuPane mp = m_WalletTabManager.getMenuPane();
				if (mp != null) {
					mp.closeVoucherPopup();
				}

				if (m_StateData != null && m_StateData.isLoggedIn()) {
					// switch to validate screen
					m_WalletTabManager.queueTransition(
														new Integer(WTabManager.M_VALIDATE));
					m_WalletTabManager.makeTransition();
				}
				else {
					m_WalletTabManager.showError("cannot validate", "not logged in",
													"you must log into your wallet first");
				}
			}
		};
		JMenuItem validateItem = new JMenuItem(m_ValidateAction);
		validateItem.setMnemonic(KeyEvent.VK_V);
		validateItem.setToolTipText("Re-validate a voucher");
		validateItem.setIcon(SparkRes.getImageIcon(SparkRes.CIRCLE_CHECK_IMAGE));
		advMenu.add(validateItem);

		// define action to split a voucher
		m_SplitAction = new AbstractAction("Split") {
			public void actionPerformed(ActionEvent actionEvent) {
				// get rid of the voucher list popup if one is displayed
				MenuPane mp = m_WalletTabManager.getMenuPane();
				if (mp != null) {
					mp.closeVoucherPopup();
				}

				if (m_StateData != null && m_StateData.isLoggedIn()) {
					// switch to split screen
					m_WalletTabManager.queueTransition(
															new Integer(WTabManager.M_SPLIT));
					m_WalletTabManager.makeTransition();
				}
				else {
					m_WalletTabManager.showError("cannot split", "not logged in",
													"you must log into your wallet first");
				}
			}
		};
		JMenuItem splitItem = new JMenuItem(m_SplitAction);
		splitItem.setMnemonic(KeyEvent.VK_S);
		splitItem.setToolTipText("Split a voucher into smaller vouchers");
		splitItem.setIcon(SparkRes.getImageIcon(SparkRes.SMALL_DOCUMENT_ADD));
		advMenu.add(splitItem);

		// define action to merge vouchers
		m_MergeAction = new AbstractAction("Merge") {
			public void actionPerformed(ActionEvent actionEvent) {
				// get rid of the voucher list popup if one is displayed
				MenuPane mp = m_WalletTabManager.getMenuPane();
				if (mp != null) {
					mp.closeVoucherPopup();
				}

				if (m_StateData != null && m_StateData.isLoggedIn()) {
					// switch to merge screen
					m_WalletTabManager.queueTransition(
															new Integer(WTabManager.M_MERGE));
					m_WalletTabManager.makeTransition();
				}
				else {
					m_WalletTabManager.showError("cannot merge", "not logged in",
													"you must log into your wallet first");
				}
			}
		};
		JMenuItem mergeItem = new JMenuItem(m_MergeAction);
		mergeItem.setMnemonic(KeyEvent.VK_M);
		mergeItem.setToolTipText("Consolidate vouchers into one");
		mergeItem.setIcon(SparkRes.getImageIcon(SparkRes.FUNNEL_DOWN_16x16));
		advMenu.add(mergeItem);

		// define action to buy tokens manually
		AbstractAction purchTok = new AbstractAction("Buy Tokens") {
			public void actionPerformed(ActionEvent actionEvent) {
				// get rid of the voucher list popup if one is displayed
				MenuPane mp = m_WalletTabManager.getMenuPane();
				if (mp != null) {
					mp.closeVoucherPopup();
				}

				if (m_StateData != null && m_StateData.isLoggedIn()) {
					// switch to token purchase screen
					m_WalletTabManager.queueTransition(
												new Integer(WTabManager.M_PURCHASE_TOKENS));
					m_WalletTabManager.makeTransition();
				}
				else {
					m_WalletTabManager.showError("cannot buy tokens",
														  "not logged in",
													"you must log into your wallet first");
				}
			}
		};
		JMenuItem purchItem = new JMenuItem(purchTok);
		purchItem.setMnemonic(KeyEvent.VK_B);
		purchItem.setToolTipText("Purchase tokens manually");
		purchItem.setIcon(SparkRes.getImageIcon(SparkRes.SMALL_AGENT_IMAGE));
		advMenu.add(purchItem);

		// define action to deal with any expired vouchers
		m_ExpiredAction = new AbstractAction("Expired Vouchers") {
			public void actionPerformed(ActionEvent actionEvent) {
				if (m_StateData != null && m_StateData.isLoggedIn()) {
					// go to expired screen
					m_WalletTabManager.queueTransition(
														new Integer(WTabManager.M_EXPIRED));
					m_WalletTabManager.makeTransition();
				}
				else {
					m_WalletTabManager.showError("cannot process expired vouchers",
														  "not logged in",
														"please log into your wallet first");
				}
			}
		};
		JMenuItem expiredItem = new JMenuItem(m_ExpiredAction);
		expiredItem.setMnemonic(KeyEvent.VK_X);
		expiredItem.setToolTipText("Deal with any expired vouchers");
		expiredItem.setIcon(SparkRes.getImageIcon(SparkRes.RED_FLAG_16x16));
		advMenu.add(expiredItem);

		// define action to close a safe
		m_CloseSafeAction = new AbstractAction("Close Safe") {
			public void actionPerformed(ActionEvent actionEvent) {
				if (m_StateData != null && m_StateData.isLoggedIn()) {
					// go to close safe screen
					m_WalletTabManager.queueTransition(
													new Integer(WTabManager.M_CLOSE_SAFE1));
					m_WalletTabManager.makeTransition();
				}
				else {
					m_WalletTabManager.showError("cannot close wallet",
														  "not logged in",
													"you must log into your wallet first");
				}
			}
		};
		JMenuItem closeItem = new JMenuItem(m_CloseSafeAction);
		closeItem.setMnemonic(KeyEvent.VK_C);
		closeItem.setToolTipText("Close out your wallet");
		closeItem.setIcon(SparkRes.getImageIcon(SparkRes.BRICKWALL_IMAGE));
		advMenu.add(closeItem);
		menuBar.add(advMenu);
		menuBar.add(Box.createHorizontalStrut(10));

		// Help menu
		JMenu helpMenu = new JMenu("Help ");
		helpMenu.setMnemonic(KeyEvent.VK_H);

		// define action for tutorials help
		m_TutorialHelpAction = new AbstractAction("Tutorials") {
			public void actionPerformed(ActionEvent actionEvent) {
				m_WalletTabManager.queueTransition(
														new Integer(WTabManager.M_TUTORIALS));
				m_WalletTabManager.makeTransition();
			}
		};
		JMenuItem tutorialsItem = new JMenuItem(m_TutorialHelpAction);
		tutorialsItem.setMnemonic(KeyEvent.VK_T);
		tutorialsItem.setToolTipText("Go to user tutorials");
		tutorialsItem.setIcon(SparkRes.getImageIcon(
																SparkRes.SMALL_DOCUMENT_VIEW));
		helpMenu.add(tutorialsItem);

		// define action for About help
		m_AboutAction = new AbstractAction("About") {
			public void actionPerformed(ActionEvent actionEvent) {
				// show version info in a popup
				JFrame vPop = new JFrame("VSC Version Info");
				vPop.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
				JTextArea version = new JTextArea(4, 50);
				version.setEditable(false);
				version.setLineWrap(true);
				// use same font as HomePane.M_DataFont
				version.setFont(new Font(Font.SANS_SERIF, Font.PLAIN, 12));
				version.setWrapStyleWord(true);
				version.setText("\n" + M_Version + "\n");
				vPop.add(version);
				vPop.setSize(new Dimension(280, 90));
				vPop.setLocationRelativeTo(menuBar);
				vPop.setVisible(true);
				registerPopup(vPop);
			}
		};
		JMenuItem aboutItem = new JMenuItem(m_AboutAction);
		aboutItem.setMnemonic(KeyEvent.VK_A);
		aboutItem.setToolTipText("Show version info");
		URL imgURL = getClass().getResource(M_InfoIconName);
		ImageIcon infoIcon = new ImageIcon(imgURL, "info icon");
		aboutItem.setIcon(infoIcon);
		helpMenu.add(aboutItem);
		menuBar.add(helpMenu);
		menuBar.add(Box.createHorizontalStrut(140));		// spacer

		// define action to log out of safe
		m_LogoutAction = new AbstractAction("Logout") {
			public void actionPerformed(ActionEvent actionEvent) {
				if (m_StateData != null && m_StateData.isLoggedIn()) {
					doLogout();

					/* Because we are now using PageKites to the OFS (and other
					 * servers), and because these grow stale a short time after
					 * :80 is pinged, and cause reconnect attempts to fail, we need
					 * to terminate the OFS connection when the wallet logs out.
					 */
					doDisconnect();
					m_WalletTabManager.queueTransition(
													new Integer(WTabManager.M_DISCONNECTED));
					m_WalletTabManager.makeTransition();
				}
				else {
					m_WalletTabManager.showInfo("You are not logged into a wallet");
				}
			}
		};
		JMenuItem logoutItem = new JMenuItem(m_LogoutAction);
		logoutItem.setToolTipText("Log out of your wallet");
		logoutItem.setOpaque(false);
		menuBar.add(Box.createHorizontalGlue());
		menuBar.add(logoutItem);
		menuBar.add(Box.createHorizontalGlue());

		setJMenuBar(menuBar);
		pack();
		setVisible(true);

		// go to appropriate first screen
		if (m_OFSConnection != null) {
			m_WalletTabManager.queueTransition(new Integer(WTabManager.M_LOGIN1));
		}
		else {
			m_WalletTabManager.queueTransition(
													new Integer(WTabManager.M_DISCONNECTED));
		}
		new Thread(this).start();
	}

	// get methods
	/**
	 * get state tracking object
	 * @return state-tracking object, of type {@link VscState}
	 */
	public VscState getVscState() { return m_StateData; }

	/**
	 * get login secrets object
	 * @return login secrets object, of type {@link VsSecrets}
	 */
	public VsSecrets getLoginSecrets() { return m_LoginSecrets; }

	/**
	 * get the OFS message listener object
	 * @return the listener
	 */
	public OFSListener getOFSListener() { return m_OFSListener; }

	/**
	 * get the PKS message listener object
	 * @return the listener
	 */
	public PKSListener getPKSListener() { return m_PKSListener; }

	/**
	 * get the web message listener object
	 * @return the listener
	 */
	public WebListener getWebListener() { return m_WebListener; }

	/**
	 * get the SVX listener object
	 * @return the listener
	 */
	public SVXListener getSVXListener() { return m_SVXListener; }

	/**
	 * get the SVM listener object
	 * @return the listener
	 */
	public SVMListener getSVMListener() { return m_SVMListener; }

	/**
	 * get the UI screen/tab manager for the wallet tab
	 * @return the tab manager for the plugin's Wallet tab
	 */
	public WTabManager getWalletTabManager() {
		return m_WalletTabManager;
	}

	/**
	 * get the UI screen/tab manager for the SVX tab
	 * @return the tab manager for the plugin's SVX tab
	 */
	public XTabManager getExchangeTabManager() {
		return m_ExchangeTabManager;
	}

	/**
	 * get the UI screen/tab manager for the SVM tab
	 * @return the tab manager for the plugin's SVM tab
	 */
	public MTabManager getMarketplaceTabManager() {
		return m_MarketplaceTabManager;
	}

	/**
	 * get the status bar
	 * @return the panel containing the status/progress bar
	 */
	public JPanel getStatusBar() { return m_StatusBar; }

	/**
	 * get the SDS listener
	 * @return the listener, of type {@link SDSListener}
	 */
	public SDSListener getSDSListener() { return m_SDSListener; }

	/**
	 * get the DHT listener
	 * @return the listener, of type {@link DHTListener}
	 */
	public DHTListener getDHTListener() { return m_DHTListener; }

	/**
	 * get the spent tokens cache
	 * @return the cache, of type {@link SpentTokenCache}
	 */
	public SpentTokenCache getTokenCache() { return m_SpentCache; }

	/**
	 * obtain the fiat asset pricing module
	 * @return the module
	 */
	public AssetPriceCache getPricingModule() { return m_PricingModule; }

	/**
	 * get the user preferences
	 * @return the object containing user preference values
	 */
	public VSPreferences getUserPreferences() { return m_UserPrefs; }

	/**
	 * get the default message wait timeout
	 * @return the timeout, in msec
	 */
	public long getTimeout() { return m_UserPrefs.getTimeout() * 1000L; }

	/**
	 * get the OFS connection
	 * @return the connection, or null if none exists
	 */
	public XMPPConnection getOFSConnection() { return m_OFSConnection; }

	/**
	 * method to obtain the log directory
	 * @return the directory where log files etc. are created
	 */
	public File getLogDir() { return m_LogDir; }

	/**
	 * return test network status
	 * @return true if we're running on a test network
	 */
	public boolean isTestNetwork() { return m_TestNet; }

	/**
	 * obtain the connect action (used in LoginPane)
	 * @return the action used to establish a connection to the selected OFS
	 */
	public AbstractAction getConnectAction() { return m_ConnectAction; }

	/**
	 * return the number of seconds since we connected to the OFS
	 * @return time since connection in seconds, or zero if not yet initialized
	 */
	public int getTimeSinceInit() {
		if (m_InitTimestamp == null) {
			return 0;
		}
		Date now = new Date();
		int duration = (int) ((now.getTime() - m_InitTimestamp.getTime()) / 1000);
		return duration;
	}

	/**
	 * obtain the current SVM plugin configuration
	 * @return the config for the currently connected (or last connected) SVM
	 */
	public SVMPlugin getSVMPlugin() { return m_SVMPlugin; }

	/**
	 * obtain the standard web message filter
	 * @return the filter which looks for website Message broadcasts
	 */
	public PacketFilter getWebMsgFilter() { return m_WebMsgFilter; }

	/**
	 * obtain the standard SVX message filter
	 * @return the filter which looks for unsolicited SVX Message broadcasts
	 */
	public PacketFilter getSVXMsgFilter() { return m_SVXMsgFilter; }

	/**
	 * obtain the standard SVM message filter
	 * @return the filter which looks for unsolicited Message broadcasts
	 */
	public PacketFilter getSVMMsgFilter() { return m_SVMMsgFilter; }


	// set methods
	/**
	 * switch the tabbed pane to select the designated Component
	 * @param comp the Component (one of the TabManager subclasses) to switch to
	 */
	public void setSelectedTab(Component comp) {
		if (comp == null) {
			return;
		}
		int tIdx = m_TabbedPane.indexOfComponent(comp);
		if (tIdx == -1) {
			Log.error("No tab for this component");
			return;
		}
		m_TabbedPane.setSelectedIndex(tIdx);
	}

	/**
	 * private helper method to establish a connection to the OFS.
	 * @param OFSdomain the XMPP domain of the OFS to be connected
	 * @param OFShost the hostname where the OFS can be found
	 * @param OFSport the port number where the OFS can be found
	 * @return a live connection to the OFS, or null if something failed
	 */
	private XMPPConnection openOFS(String OFSdomain,
											 String OFShost,
											 int OFSport)
	{
		if (OFSdomain == null || OFSdomain.isEmpty()) {
			Log.error("Missing OFS domain designation, cannot connect");
			return null;
		}
		if (OFShost == null || OFShost.isEmpty()) {
			Log.error("Missing OFS host designation, cannot connect");
			return null;
		}
		if (OFSport <= 0) {
			Log.error("Missing OFS port designation, cannot connect");
			return null;
		}

		// on an OFS connection, always attempt port 80 pagekite ping first
		port80KitePing(OFShost);

		ConnectionConfiguration cc
			= new ConnectionConfiguration(OFShost, OFSport, OFSdomain);
		cc.setCompressionEnabled(true);
		cc.setNotMatchingDomainCheckEnabled(true);
		cc.setRosterLoadedAtLogin(false);
		cc.setSecurityMode(ConnectionConfiguration.SecurityMode.required);
		cc.setSelfSignedCertificateEnabled(true);
		cc.setSendPresence(false);
		final XMPPConnection ofsConn = new XMPPConnection(cc);
		m_OFSConnectStat = false;
		Runnable ofsConnector = new Runnable() {
			public void run() {
				try {
					ofsConn.connect();
					ofsConn.loginAnonymously();
					if (ofsConn.isUsingTLS()) {
						m_OFSConnectStat = true;
					}
					else {
						// abandon unsecure connection
						ofsConn.disconnect();
					}
				}
				catch (XMPPException xe) {
					Log.error("Could not connect to OFS", xe);
					m_WalletTabManager.showError("OFS connect fail",
															xe.getMessage());
				}
			}
		};
		Thread doConnect = new Thread(ofsConnector);

		// allow 10 secs for OFS connect and anonymous login
		doConnect.start();
		try {
			doConnect.join(10000L);
		}
		catch (InterruptedException ie) {
			if (!m_OFSConnectStat) {
				// give up
				doConnect.interrupt();
			}
		}
		if (m_OFSConnectStat && ofsConn.isConnected()) {
			m_InitTimestamp = new Date();
			return ofsConn;
		}
		return null;
	}

	/**
	 * helper method to find the address of the DHT and connect to it.
	 * (tries multiple paths in sequence if they exist)
	 * @return a live connection to the DHT, or null if something failed
	 */
	public XMPPConnection openDHT() {
		// see if we already have one
		XMPPConnection retConn = m_DHTListener.getDHTConnection();
		if (retConn != null)  {
			return retConn;
		}
		ArrayList<InetSocketAddress> dhtList = m_LoginSecrets.getDHTNodes();
		if (dhtList.isEmpty()) {
			Log.error("No DHTs recorded!");
			m_WalletTabManager.showError("No DHT set!");
			return retConn;
		}

		// reduce connect times (on connect, Smack seems to use 3x reply time)
		int orgTimeout = SmackConfiguration.getPacketReplyTimeout();
		SmackConfiguration.setPacketReplyTimeout(10000);

		// take the list in order, favoring a path with the same domain as the OFS
		String dhtDomain = m_LoginSecrets.getDHTDomain();
		Iterator<InetSocketAddress> dhti = dhtList.iterator();
		InetSocketAddress usedAddr = null;
		while (dhti.hasNext()) {
			InetSocketAddress dhtAddr = dhti.next();

			// make a connection to the DHT Openfire server
			ConnectionConfiguration cc
				= new ConnectionConfiguration(dhtAddr.getHostString(),
														dhtAddr.getPort(),
														dhtDomain);
			cc.setSecurityMode(ConnectionConfiguration.SecurityMode.required);
			XMPPConnection dhtConn = new XMPPConnection(cc);

			// if we're not on the first iteration, try a pagekite 80 ping first
			boolean pinged = false;
			if (dhtList.indexOf(dhtAddr) != 0) {
				if (!port80KitePing(dhtAddr.getHostString())) {
					Log.error("Ping " + dhtAddr.getHostString() + ":80 failed");
					// presumably, connection also won't work
					continue;
				}
				pinged = true;
			}
			try {
				dhtConn.connect();
			}
			catch (XMPPException xe) {
				Log.error("Could not connect to DHT " + dhtDomain + " at "
							+ dhtAddr.getHostString() + ":" + dhtAddr.getPort());

				// attempt retry after pinging pagekite 80 port
				if (!pinged && port80KitePing(dhtAddr.getHostString())) {
					try {
						dhtConn.connect();
					}
					catch (XMPPException xe2) {
						Log.error("Also failed after :80 ping");
						continue;
					}
				}
				else {
					continue;
				}
			}
			// perform anon login
			try {
				dhtConn.loginAnonymously();
				retConn = dhtConn;
			/*
				Log.debug("Successful connect to DHT " + dhtDomain + " at path "
							+ dhtAddr.getHostString() + ":" + dhtAddr.getPort());
			 */
				usedAddr = dhtAddr;
				break;
			}
			catch (XMPPException xe) {
				Log.error("Could not login anon to DHT " + dhtDomain + " at path "
							+ dhtAddr.getHostString() + ":" + dhtAddr.getPort(), xe);
				dhtConn.disconnect();
			}
		}
		SmackConfiguration.setPacketReplyTimeout(orgTimeout);
		if (retConn == null) {
			m_WalletTabManager.showError("DHT connect failure");
			return retConn;
		}

		// record connection with the listener so it can be closed later
		m_DHTListener.setDHTConnection(retConn);

		// if needed, resort the DHT and SDS list to the connected hostname is 1st
		String connHost = retConn.getHost();
		if (dhtList.indexOf(usedAddr) != 0) {
			reorderGateways(dhtList, connHost);
			// do the same for SDS if we don't have a connection yet
			if (m_SDSListener.getSDSConnection() == null) {
				reorderGateways(m_LoginSecrets.getSDSGateways(), connHost);
			}
		}
		return retConn;
	}
	
	/**
	 * helper method to find the address of the SDS and connect to it.
	 * (tries multiple paths in sequence if they exist)
	 * @return a live connection to the SDS, or null if something failed
	 */
	public XMPPConnection openSDS() {
		// see if we already have one
		XMPPConnection retConn = m_SDSListener.getSDSConnection();
		if (retConn != null)  {
			return retConn;
		}
		ArrayList<InetSocketAddress> sdsList = m_LoginSecrets.getSDSGateways();
		if (sdsList.isEmpty()) {
			Log.error("No SDSs recorded!");
			m_WalletTabManager.showError("No SDS set!");
			return retConn;
		}

		// reduce connect times (on connect, Smack seems to use 3x reply time)
		int orgTimeout = SmackConfiguration.getPacketReplyTimeout();
		SmackConfiguration.setPacketReplyTimeout(10000);

		// take the list in order, favoring a path with the same domain as the OFS
		String sdsDomain = m_LoginSecrets.getSDSDomain();
		Iterator<InetSocketAddress> sdsi = sdsList.iterator();
		InetSocketAddress usedAddr = null;
		while (sdsi.hasNext()) {
			InetSocketAddress sdsAddr = sdsi.next();

			// make a connection to the SDS Openfire server
			ConnectionConfiguration cc
				= new ConnectionConfiguration(sdsAddr.getHostString(),
														sdsAddr.getPort(),
														sdsDomain);
			cc.setSecurityMode(ConnectionConfiguration.SecurityMode.required);
			XMPPConnection sdsConn = new XMPPConnection(cc);

			// if we're not on the first iteration, try a pagekite 80 ping first
			boolean pinged = false;
			if (sdsList.indexOf(sdsAddr) != 0) {
				if (!port80KitePing(sdsAddr.getHostString())) {
					Log.error("Ping " + sdsAddr.getHostString() + ":80 failed");
					// presumably, connection also won't work
					continue;
				}
				pinged = true;
			}
			try {
				sdsConn.connect();
			}
			catch (XMPPException xe) {
				Log.error("Could not connect to SDS " + sdsDomain + " at "
							+ sdsAddr.getHostString() + ":" + sdsAddr.getPort());

				// attempt retry after pinging pagekite 80 port
				if (!pinged && port80KitePing(sdsAddr.getHostString())) {
					try {
						sdsConn.connect();
					}
					catch (XMPPException xe2) {
						Log.error("Also failed after :80 ping");
						continue;
					}
				}
				else {
					continue;
				}
			}
			// perform anon login
			try {
				sdsConn.loginAnonymously();
				retConn = sdsConn;
			/*
				Log.debug("Successful connect to SDS " + sdsDomain + " at path "
							+ sdsAddr.getHostString() + ":" + sdsAddr.getPort());
			 */
				usedAddr = sdsAddr;
				break;
			}
			catch (XMPPException xe) {
				Log.error("Could not login anon to SDS " + sdsDomain + " at path "
							+ sdsAddr.getHostString() + ":" + sdsAddr.getPort(), xe);
				sdsConn.disconnect();
			}
		}
		SmackConfiguration.setPacketReplyTimeout(orgTimeout);
		if (retConn == null) {
			m_WalletTabManager.showError("SDS connect failure");
			return retConn;
		}

		// record connection with the listener so it can be closed later
		m_SDSListener.setSDSConnection(retConn);

		// if needed, resort the SDS and DHT list to the connected hostname is 1st
		String connHost = retConn.getHost();
		if (sdsList.indexOf(usedAddr) != 0) {
			reorderGateways(sdsList, connHost);
			// do the same for DHT if we don't have a connection yet
			if (m_DHTListener.getDHTConnection() == null) {
				reorderGateways(m_LoginSecrets.getDHTNodes(), connHost);
			}
		}
		return retConn;
	}

	/**
	 * utility method to log out of a wallet (assumes login process completed)
	 */
	public void doLogout() {
		m_WalletTabManager.setCursor(
										Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
		m_StateData.setLoggedIn(false);
		m_SpentCache.purgeSpentTokens();
		m_DHTListener.closeDHTConnection();
		m_DHTListener.getDHTrecs().clear();
		m_SDSListener.closeSDSConnection();
		m_PKSListener.closePKSConnection();
		m_WalletTabManager.setCursor(null);

		// remove any popup windows we might have
		for (JFrame popup : m_PopupList) {
			popup.dispose();
		}
		m_PopupList.clear();

		// reset states on marketplace and exchange tabs
		m_ExchangeTabManager.queueTransition(
												new Integer(XTabManager.M_LOGOUT_EXCHANGE));
		m_ExchangeTabManager.makeTransition();
		m_MarketplaceTabManager.queueTransition(
												new Integer(MTabManager.M_LOGOUT_MARKET));
		m_MarketplaceTabManager.makeTransition();
		resetSecrets();
		// forget the VPs asset/unit/token configuration
		VPAssetConfig.m_AssetConfig.clear();

		// set the main tab state to wallet tab manager
		setSelectedTab(m_WalletTabManager);
		System.gc();
	}

	/**
	 * handle disconnecting a live OFS connection and associated tasks
	 */
	private void doDisconnect() {
		if (m_OFSConnection == null) {
			return;
		}
		// remove listeners
		m_OFSConnection.removePacketListener(m_OFSListener);
		m_OFSConnection.removePacketListener(m_SVMListener);
		m_OFSConnection.removePacketListener(m_SVXListener);
		m_OFSConnection.removePacketListener(m_WebListener);
		m_OFSConnection.removeConnectionListener(m_OFSConnectListener);
		// disconnect XMPP
		m_OFSConnection.disconnect();
		m_OFSConnection = null;
		// kill price fetching, which requires OFS connection
		setDefaultStatusBar();
		m_PricingModule.cancelPriceFetching();
		m_PricingModule = null;
	}

	/**
	 * set the default status bar contents, based on connection status
	 */
	public void setDefaultStatusBar() {
		String connectTxt = null;
		int strut = 0;
		if (m_OFSConnection != null && m_OFSConnection.isConnected()) {
			connectTxt = " Connected to OFS gateway - Ready";
			strut = 260;
		}
		else {
			connectTxt = " Not connected";
			strut = 400;
		}
		final JLabel conLab = new JLabel(connectTxt);
		conLab.setFont(m_OFSConnectListener.M_LabFont);
		final Component strutBox = Box.createHorizontalStrut(strut);
		m_StatusBar.removeAll();
		m_StatusBar.add(conLab);
		m_StatusBar.add(strutBox);
		m_StatusBar.invalidate();
		m_StatusBar.revalidate();
		m_StatusBar.repaint();
		invalidate();
		repaint();
	}

	/**
	 * Method to implement Runnable, so that we can invoke UI elements in an
	 * asynchronous manner, without risking the termination or hijacking of
	 * this thread (probably an event thread) by Swing.
	 */
	public void run() {
		m_WalletTabManager.makeTransition();
	}

	/**
	 * register a new popup window, for auto-closure during logout
	 * @param win the popup window to be recorded
	 */
	public void registerPopup(JFrame win) {
		if (win != null) {
			m_PopupList.add(win);
		}
	}

	/**
	 * remove a popup window, in the event it was already disposed of
	 * @param win the popup window to be deregistered
	 */
	public void deregisterPopup(JFrame win) {
		if (win != null) {
			m_PopupList.remove(win);
		}
	}

	/**
	 * set test network status
	 * @param stat true if we're running on a test network
	 */
	public void setTestNetwork(boolean stat) { m_TestNet = stat; }

	/**
	 * do discovery of Components at the OFS
	 * @return true on successful send of disco request.
	 * (NB: reply goes to OFSListener)
	 */
	public boolean doDisco() {
		if (m_OFSConnection == null || !m_OFSConnection.isConnected()) {
			Log.error("Cannot do disco without connection");
			return false;
		}
		DiscoverItems disco = new DiscoverItems();
		disco.setTo(m_OFSConnection.getServiceName());
		disco.setFrom(m_OFSConnection.getUser());
		disco.setType(IQ.Type.GET);
		String discoId = "disco_" + m_StateData.getNextOFSid();
		disco.setPacketID(discoId);

		// register a collector for the reply
		PacketIDFilter pFilter = new PacketIDFilter(discoId);
		PacketCollector repColl = m_OFSConnection.createPacketCollector(pFilter);

		// send request to server
		m_OFSConnection.sendPacket(disco);
		m_StateData.setLastActivity(disco);

		// wait for reply or timeout
		IQ reply = (IQ) repColl.nextResult(getTimeout());
		if (reply == null) {
			Log.error("Timeout on discovery request");
			return false;
		}
		DiscoverItems discoItems = (DiscoverItems) reply;
		Iterator<DiscoverItems.Item> dIt = discoItems.getItems();
		boolean foundSVX = false;
		ArrayList<String> svxIDs = new ArrayList<String>();
		ArrayList<String> svmIDs = new ArrayList<String>();
		boolean foundSVM = false;
		while (dIt.hasNext()) {
			DiscoverItems.Item di = dIt.next();
			/* NB: these strings must match the value of M_Description in the
			 * 	 code for the OFS's Components.
			 */
			String name = di.getName();
			if (name.startsWith("Silent Vault Exchange (SVX) Interface Component"))
			{
				foundSVX = true;
				// add the ID to the list
				int delimiterIdx = name.indexOf("::");
				if (delimiterIdx != -1) {
					String svxId = name.substring(delimiterIdx+2);
					svxIDs.add(svxId);
				}
			}
			else if (name.startsWith(
									"Silent Vault Market (SVM) Interface Component"))
			{
				foundSVM = true;
				// add the ID to the list
				int delimiterIdx = name.indexOf("::");
				if (delimiterIdx != -1) {
					String svmId = name.substring(delimiterIdx+2);
					svmIDs.add(svmId);
				}
			}
		}
		if (foundSVX) {
			// tell the ConnectPane about the SVX list
			m_ExchangeTabManager.getConnectPane().recordSVXList(svxIDs);
		}
		if (foundSVM) {
			// tell the LoginPane about the SVM list
			m_MarketplaceTabManager.getLoginPane().recordSVMList(svmIDs);
		}
		return true;
	}

	/**
	 * method to attempt a connect to a PageKite'd XMPP host prior to connection
	 * @param xmppHost the hostname to connect to
	 * @return true on success
	 */
	public boolean port80KitePing(String xmppHost) {
		/* The following is a hack for supporting OFS hosts which are PageKites.
		 * A PageKite requires an open HTTP connection from the same incoming IP
		 * in order to proxy an encrypted connection across a raw port.  We
		 * therefore open '/' on the host first, before initiating XMPP connect.
		 * However, in the event that the OFShost isn't a PageKite and doesn't
		 * have anything listening on port 80, we will not let a failure abort
		 * our connect attempt.
		 */
		boolean ret = false;
		try {
			URL pkURL = new URL("http", xmppHost, 80, "/");
			URLConnection pkConn = pkURL.openConnection();
			HttpURLConnection preConn = (HttpURLConnection) pkConn;
			preConn.setRequestMethod("GET");
			preConn.setAllowUserInteraction(false);
			preConn.setDoOutput(true);
			preConn.setConnectTimeout(10000);
			preConn.setReadTimeout(10000);
			preConn.connect();
			Object junk = preConn.getContent();
			preConn.disconnect();
			ret = true;
		}
		catch (MalformedURLException mue) {
			Log.error("Could not construct URL for pre-connect step", mue);
		}
		catch (ProtocolException pe) {
			Log.error("Could not connect to URL for pre-connect step", pe);
		}
		catch (IOException ioe) {
			Log.error("Could not connect to URL for pre-connect step");
		}
		return ret;
	}

	/**
	 * method to reorder the gateways on a list, based on which should be first
	 * @param gtwyList the list of gateways to be sorted
	 * @param firstHost the hostname which should become first on the list
	 */
	public void reorderGateways(ArrayList<InetSocketAddress> gtwyList,
										 String firstHost)
	{
		if (gtwyList == null || gtwyList.isEmpty()) {
			return;
		}

		// find the indicated host
		InetSocketAddress firstEntry = null;
		for (InetSocketAddress isa : gtwyList) {
			if (isa.getHostString().equals(firstHost)) {
				firstEntry = isa;
				break;
			}
		}
		if (firstEntry != null) {
			// remove from list
			gtwyList.remove(firstEntry);
			// add it back at index 0
			gtwyList.add(0, firstEntry);
		}
		else {
			Log.error("Host " + firstHost + " not found on gtwyList");
		}
	}

	/**
	 * reset the SVM listener
	 * @param newListener the replacement (override) listener to use
	 */
	public void setSVMListener(SVMListener newListener) {
		// remove any packet listener which went to the old one
		if (m_OFSConnection != null && m_SVMListener != null) {
			m_OFSConnection.removePacketListener(m_SVMListener);
		}

		// if null, go back to the default
		if (newListener == null) {
			m_SVMListener = new SVMListener(this);
		}
		else {
			// use what was passed
			m_SVMListener = newListener;
		}

		// add packet listener pointing at the new one
		if (m_OFSConnection != null) {
			m_OFSConnection.addPacketListener(m_SVMListener, m_SVMMsgFilter);
		}
	}

	/**
	 * reset the SVM tab manager
	 * @param tabManager the new tab manager to set
	 */
	public void setSVMTabManager(MTabManager tabManager) {
		if (tabManager != null) {
			m_MarketplaceTabManager = tabManager;
			m_TabbedPane.setComponentAt(2, m_MarketplaceTabManager);
		}
	}

	/**
	 * set the SVM plugin configuration
	 * @param svmPlug the new plugin to use
	 */
	public void setSVMPlugin(SVMPlugin svmPlug) {
		if (svmPlug != null) {
			m_SVMPlugin = svmPlug;
		}
	}

	// END methods
}
