/*
 * last modified---
 * 	07-12-13/jet/ add addWebMessage()
 * 	06-05-13/jet/ SilentVault client version
 * 	04-19-10/kbw/ stand-alone client version
 * 	08-27-09/kbw/ new
 *
 * purpose---
 * 	listen for messages from websites, which are incoming unsolicited
 *
 * notes---
 * 	see underlying messages in web_vsc.x
 */

package org.silentvault.client;

import org.jivesoftware.smack.PacketListener;
import org.jivesoftware.smack.packet.Packet;
import org.jivesoftware.smack.packet.Message;
import org.jivesoftware.smack.packet.PacketExtension;

import java.util.ArrayList;


/**
 * class to handle processing of incoming website order specification messages,
 * in the jabber:message:voucher-safe#web-vsc namespace.
 *
 * @author Kevin Wilkerson
 */
public final class WebListener implements PacketListener {
	// BEGIN data members
	/**
	 * a list of received order messages from websites
	 */
	private ArrayList<WebMessage>	m_OrderMessages;

	// END data members

	// BEGIN methods
	/**
	 * constructor
	 */
	public WebListener() {
		m_OrderMessages = new ArrayList<WebMessage>(10);
	}

	// method required to implement the PacketListener interface
	/**
	 * deal with an incoming message from a website
	 * @param packet the packet (will really be a WebMessage)
	 */
	public void processPacket(Packet packet) {
		// verify we got what's expected
		if (packet instanceof Message) {
			Message msg = (Message) packet;
			PacketExtension pe = msg.getExtension("web_vsc",
													"jabber:message:voucher-safe#web-vsc");
			if (pe == null) {
				// must be a normal chat message
				Log.debug("Normal chat message received, ignoring");
				return;
			}
			WebMessage webMsg = (WebMessage) pe;
			webMsg.setTo(msg.getTo());
			webMsg.setFrom(msg.getFrom());
			webMsg.setType(msg.getType());
			webMsg.setSubject(msg.getSubject());
			webMsg.setBody(msg.getBody());
			if (!webMsg.getInfo().isEmpty()) {
				// looks like a valid encrypted message, add it to list
				m_OrderMessages.add(webMsg);
			}
		}
		// not interested in IQs or Presence
	}

	/**
	 * get the list of received website messages
	 * @return the list
	 */
	public ArrayList<WebMessage> getWebsiteMessages() { return m_OrderMessages; }

	/**
	 * remove a processed message from the list of received website messages
	 * @param webMsg the message to remove
	 */
	public void purgeWebMessage(WebMessage webMsg) {
		// does nothing if it wasn't there to begin with
		m_OrderMessages.remove(webMsg);
	}

	/**
	 * manually add a WebMessage
	 * @param webMsg the message, constructed programmatically
	 */
	public void addWebMessage(WebMessage webMsg) {
		if (webMsg != null) {
			m_OrderMessages.add(webMsg);
		}
	}

	/**
	 * finalize object when garbage-collected, for data security
	 */
	@Override
	protected void finalize() throws Throwable {
		try {
			if (m_OrderMessages != null && !m_OrderMessages.isEmpty()) {
				m_OrderMessages.clear();
			}
		} finally {
			super.finalize();
		}
	}

	// END methods
}
