/*
 * last modified---
 * 	06-05-13/jet/ SilentVault client version
 * 	04-19-10/kbw/ stand-alone client version
 * 	08-27-09/kbw/ new
 *
 * purpose---
 * 	Parse {@link WebMessage} packet extensions.  Parses the web_vsc
 * 	sub-document and creates an instance of {@link WebMessage}.
 *
 * notes---
 * 	see underlying message definition in web_vsc.x
 */

package org.silentvault.client;

import org.xmlpull.v1.XmlPullParser;

import org.jivesoftware.smack.packet.PacketExtension;
import org.jivesoftware.smack.packet.Message;
import org.jivesoftware.smack.packet.DefaultPacketExtension;
import org.jivesoftware.smack.provider.PacketExtensionProvider;


/**
 * Implements the PacketExtensionProvider interface to receive and parse
 * incoming messages from a website, according to the web_vsc element of
 * the jabber:message:voucher-safe#web-vsc namespace.  Creates a
 * {@link WebMessage} to hold the parsed message packet extension.
 *
 * @author Kevin Wilkerson
 */
public class WebMessageProvider implements PacketExtensionProvider {
	// members
	private static final String	M_EndTag = "web_vsc";
	
	// methods
	/**
	 * constructor
	 */
	public WebMessageProvider() {
	}

	/**
	 * parse the packet extension
	 * @param parser the parser, positioned at the start of the extension
	 * @return the extension element
	 */
	public PacketExtension parseExtension(XmlPullParser parser) throws Exception
	{
		WebMessage webMsg = new WebMessage();
		// NB: these values should be web_vsc, jabber:message:voucher-safe#web-vsc
		webMsg.addExtension(new DefaultPacketExtension(parser.getName(),
																	  parser.getNamespace()));
		webMsg.setType(Message.Type.normal);

		// at this stage, the extension will be encrypted and contain only:
		// <info>base64-encoded 3DES-encrypted string</info>

		String info = "";
		int eventType = parser.next();
		if (eventType == XmlPullParser.START_TAG) {
			if (parser.getName().equals("info")) {
				info = parser.nextText().trim();
			}
			else {
				throw new Exception("Unexpected element, " + parser.getName());
			}
		}
		else {
			throw new Exception("missing " + M_EndTag + " info");
		}
		// we do not decode or decrypt, yet
		webMsg.setInfo(info);

		// get end tag
		eventType = parser.next();
		if (eventType == XmlPullParser.END_TAG) {
			if (!parser.getName().equals(M_EndTag)) {
				throw new Exception("missing /" + M_EndTag);
			}
		}
		else {
			throw new Exception("Spurious data after <info/> element");
		}
		return webMsg;
	}
}
