/*
 * last modified---
 * 	08-28-13/jet/ credit btc-e.com for LTC prices
 * 	06-05-13/jet/ SilentVault client version
 * 	03-11-12/kbw/ cite bitcoincharts.com as BTC price source
 * 	06-13-11/kbw/ remove lingo about all transactions being in GAU
 * 	03-03-11/kbw/ cosmetic changes for non-GAU vouchers
 * 	09-02-10/kbw/ add screen label
 * 	08-16-10/kbw/ remove superfluous return button
 * 	04-26-10/kbw/ flesh out
 * 	03-29-10/kbw/ new (stub)
 *
 * purpose---
 * 	provide a UI widget to handle display of underlying asset prices
 */

package org.silentvault.client.ui.wallet;

import javax.swing.*;
import java.awt.event.ActionEvent;
import java.awt.Color;

import org.silentvault.client.WalletClient;
import org.silentvault.client.AssetPriceCache;
import org.silentvault.client.XMLVoucher;

import java.text.NumberFormat;
import java.text.DecimalFormat;


/**
 * Extends WorkPane to provide a screen for showing backing asset prices.
 *
 * @author Kevin Wilkerson
 */
public final class PricesPane extends WorkPane {
	// BEGIN data members
	/**
	 * the action to return to the main menu
	 */
	private AbstractAction		m_DoneAction;

	// END data members
	
	// BEGIN methods
	/**
	 * constructor
	 * @param plugin the owning plugin or application
	 * @param home the HomePane in which we're installed
	 * @param mgr the tab manager that creates us
	 */
	public PricesPane(WalletClient plugin, HomePane home, WTabManager mgr) {
		super(plugin, home, mgr);
		m_UserInstructs = "This list of gold prices in various other "
								+ "currencies is provided solely for reference and "
								+ "comparison.<br/><br/>"
								+ "Please note that all voucher transactions are "
								+ "carried out in the units of the voucher.  "
								+ "Equivalent prices are subject to change in real "
								+ "time as world metals and currency markets "
								+ "fluctuate.";
		m_WorkArea.setBorder(null);
	}

	// method required to extend WorkPane
	/**
	 * method to prepare display for this sub-class
	 */
	public void prepDisplay() {
		// add the return to overview button as the sole contents of toolbar
		m_WorkBar.removeAll();
		m_WorkBar.add(m_MainButton);
		m_WorkBar.setVisible(true);

		// make header visible and clear buttons
		buildValueHeader();
		m_ButtonPane.removeAll();

		// set screen label
		buildScreenLabel("Gold Market Spot Prices");

		// build a simple XHTML table using the AssetPriceCache's data
		AssetPriceCache apc = m_Plugin.getPricingModule();
		StringBuilder priceTab = new StringBuilder(4096);
		priceTab.append("<html>\n<body>\n");
		priceTab.append("<table border=\"2\" ");
		priceTab.append("cellspacing=\"10\" cellpadding=\"5\">\n");
		priceTab.append("<caption><b>Gold price quoted in various other ");
		priceTab.append("currencies</b></caption>\n<thead>\n<tr>\n");
		priceTab.append("<th align=\"left\">Currency</th>\n");
		priceTab.append("<th align=\"center\">Code</th>\n");
		priceTab.append("<th align=\"right\">Price per gram (GAU)</th>\n");
		priceTab.append("<th align=\"right\">Price per troy ounce (XAU)</th>\n");
		priceTab.append("</tr>\n</thead>\n<tbody>\n");
		Double price = null;
		DecimalFormat df = (DecimalFormat) NumberFormat.getInstance();
		df.setMaximumFractionDigits(XMLVoucher.M_Precision);
		df.setMinimumFractionDigits(XMLVoucher.M_Precision);
		df.setGroupingUsed(false);

		// force a price fetch to ensure we have current data
		price = apc.getCurrPrice("USD");
		for (String cCode : apc.getSupportedCurrencies()) {
			priceTab.append("<tr>\n<td align=\"left\">");
			priceTab.append(apc.getCurrencyName(cCode));
			priceTab.append("</td>\n<td align=\"center\">" + cCode);
			priceTab.append("</td>\n<td align=\"right\">");
			price = apc.getCurrPrice(cCode);
			if (price == null) {
				price = new Double("0.0");
			}
			priceTab.append(df.format(price.doubleValue()) + "</td>\n");
			double ozPrice = price.doubleValue() * XMLVoucher.M_GramsPerOz;
			priceTab.append("<td align=\"right\">" + df.format(ozPrice));
			priceTab.append("</td>\n</tr>\n");
		}
		priceTab.append("</tbody>\n<tfoot>\n");
		priceTab.append("<tr><th colspan=\"4\">Metals and currency prices ");
		priceTab.append("courtesy of <i>dgcsc.org</i> and <i>kitco.com</i>.");
		priceTab.append("<br/>");
		priceTab.append("Bitcoin prices courtesy of <i>bitcoincharts.com</i>.");
		priceTab.append("<br/>Litecoin prices courtesy of <i>btc-e.com</i>.");
		priceTab.append("</th></tr>\n</tfoot>\n</table>\n");
		priceTab.append("</body>\n</html>\n");

		// create a viewer and load it with the table
		JEditorPane jep = new JEditorPane();
		jep.setContentType("text/html");
		jep.setEditable(false);
		jep.setText(priceTab.toString());
		jep.setBorder(null);
		m_WorkArea.setViewportView(jep);

		// count this as user activity, even though it's not communication
		// with a partner node

		m_Plugin.getVscState().setLastActivity(null);
	}

	// END methods
}
