/*
 * last modified---
 * 	07-17-15/jet/ let switch to prices screen cause fetch of latest data
 * 	03-17-15/jet/ allow for case where SVM needs to receive escrow confirm
 * 	08-19-14/jet/ back merge changes from SV Spark client plugin
 * 	09-12-13/jet/ change retrieveNetworkConfig() to await synchronous reply
 * 	08-09-13/jet/ force logout if no SDS connection available
 * 	07-12-13/jet/ check with XTabManager on main returns to continue escrow
 * 	06-05-13/jet/ SilentVault client version
 * 	01-22-13/kbw/ remove trigger of housekeeping based on expired vouchers
 * 	04-18-11/kbw/ use ConcurrentHashMap for raw DHT records
 * 	11-22-10/kbw/ display actual tutorials page in browser
 * 	09-23-10/kbw/ add capability to return to main after marking read on DHT
 * 	09-17-10/kbw/ add screen states M_SCREEN_HELP and M_TUTORIALS
 * 	08-15-10/kbw/ branch to tokens purchase before operations, not after
 * 	08-12-10/kbw/ make retrieveNetworkConfig() asynchronous
 * 	04-15-10/kbw/ add setPendingTokens() and M_PENDING_TOKENS mechanism
 * 	03-26-10/kbw/ rework for new UI spec
 *
 *	purpose---
 *		provide a manager for the contents of the application's content pane
 */

package org.silentvault.client.ui.wallet;

import javax.swing.*;
import java.awt.Cursor;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.Desktop;

import org.jivesoftware.smack.XMPPConnection;
import org.jivesoftware.smack.ConnectionConfiguration;
import org.jivesoftware.smack.XMPPException;
import org.jivesoftware.smack.packet.IQ;
import org.jivesoftware.smack.PacketCollector;
import org.jivesoftware.smack.filter.PacketIDFilter;
import org.jivesoftware.smack.util.StringUtils;

import org.silentvault.client.*;
import org.silentvault.client.ui.*;
import org.silentvault.client.ui.svx.*;
import org.silentvault.client.ui.svm.*;

import java.util.Hashtable;
import java.util.ArrayList;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ConcurrentHashMap;
import java.net.URI;
import java.net.URISyntaxException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.io.IOException;


/**
 * This class is used to manage the graphical contents of the content pane.
 * It handles creating and destroying
 * the embedded pane subclasses and manages state transitions between them.
 * It also provides utility functions such as error popups and wait cursors.
 *
 * @author Justin Turrell
 */
public class WTabManager extends TabManager {
	// BEGIN data members
	// possible states we can do
	/**
	 * initialize first VS login screen
	 */
	public static final int	M_LOGIN1 = 1;
	/**
	 * proceed to second VS login screen
	 */
	public static final int	M_LOGIN2 = 2;
	/**
	 * initialize first VS creation screen
	 */
	public static final int	M_NEW_SAFE1 = 3;
	/**
	 * proceed to second VS creation screen
	 */
	public static final int	M_NEW_SAFE2 = 4;
	/**
	 * proceed to third VS creation screen
	 */
	public static final int	M_NEW_SAFE3 = 5;
	/**
	 * proceed to fourth VS creation screen
	 */
	public static final int	M_NEW_SAFE4 = 6;
	/**
	 * move to home state, following login
	 */
	public static final int	M_HOME_INIT = 7;
	/**
	 * redisplay the voucher panel after changes
	 */
	public static final int	M_HOME_VOUCH_REGEN = 8;
	/**
	 * redisplay the token panel after changes
	 */
	public static final int	M_HOME_TOKEN_REGEN = 9;
	/**
	 * (re-)display the DHT item list
	 */
	public static final int	M_WORK_DHT_REGEN = 10;
	/**
	 * (re-)display the receipt item list
	 */
	public static final int	M_WORK_RECEIPT_REGEN = 11;
	/**
	 * (re-)display the payment item list
	 */
	public static final int	M_WORK_PAYMENT_REGEN = 12;
	/**
	 * initialize first VS recovery screen
	 */
	public static final int	M_RECOV1 = 13;
	/**
	 * proceed to second VS recovery screen
	 */
	public static final int	M_RECOV2 = 14;
	/**
	 * proceed to third VS recovery screen
	 */
	public static final int	M_RECOV3 = 15;
	/**
	 * initialize first VS closure screen
	 */
	public static final int M_CLOSE_SAFE1 = 16;
	/**
	 * proceed to second VS closure screen
	 */
	public static final int M_CLOSE_SAFE2 = 17;
	/**
	 * log out and return to login screen
	 */
	public static final int M_LOGOUT_SAFE = 18;
	/**
	 * initialize user preference UI screen
	 */
	public static final int M_PREFERENCES = 19;
	/**
	 * return to home state from preference screen (while logged in)
	 */
	public static final int M_HOME_RETURN = 20;
	/**
	 * display disconnected screen
	 */
	public static final int	M_DISCONNECTED = 21;
	/**
	 * go to payments screen
	 */
	public static final int M_PAYMENTS1 = 22;
	/**
	 * go to receipt listing screen
	 */
	public static final int M_RECEIPT_LIST = 23;
	/**
	 * go to voucher validation screen
	 */
	public static final int M_VALIDATE = 24;
	/**
	 * go to voucher split screen
	 */
	public static final int M_SPLIT = 25;
	/**
	 * go to voucher merge screen
	 */
	public static final int M_MERGE = 26;
	/**
	 * go to prices display screen
	 */
	public static final int	M_PRICES = 27;
	/**
	 * return to the main menu screen
	 */
	public static final int	M_GO_MAIN_MENU = 28;
	/**
	 * go to the incoming payment screen
	 */
	public static final int	M_INCOMING_PAYMENT = 29;
	/**
	 * go to the incoming receipt screen
	 */
	public static final int	M_INCOMING_RECEIPT = 30;
	/**
	 * go to the unclaimed payment screen
	 */
	public static final int	M_UNCLAIMED = 31;
	/**
	 * go to the expired voucher deletion screen
	 */
	public static final int	M_EXPIRED = 32;
	/**
	 * return to housekeeping after processing an item
	 */
	public static final int M_HOUSEKEEPING_RETURN = 33;
	/**
	 * go to display of a payment confirmation
	 */
	public static final int	M_CONFIRMATION = 34;
	/**
	 * go buy tokens
	 */
	public static final int	M_PURCHASE_TOKENS = 35;
	/**
	 * go back to whatever screen we were on when we had to go purchase tokens
	 */
	public static final int	M_PENDING_TOKENS = 36;
	/**
	 * display tutorials in a popup window
	 */
	public static final int	M_TUTORIALS = 37;


	// display panes we control
	/**
	 * the login pane, which handles both phases of logins and associated data
	 */
	private LoginPane			m_LoginPane;

	/**
	 * the home pane, which handles display of voucher safe contents and controls
	 */
	private HomePane			m_HomePane;

	/**
	 * the new safe pane, which handles all phases of creating a safe
	 */
	private NewSafePane		m_NewSafePane;

	/**
	 * the recovery pane, which handles recovering lost safe data
	 */
	private RecoveryPane		m_RecoveryPane;

	/**
	 * the close safe pane, which handles all phases of closing a safe
	 */
	private CloseSafePane	m_CloseSafePane;

	/**
	 * the housekeeping pane, which handles items retrieved from the DHT
	 */
	private HousekeepingPane	m_HousekeepingPane;

	/**
	 * the main menu pane, which is the home state after login and housekeeping
	 */
	private MenuPane			m_MenuPane;

	/**
	 * the receipts pane, which handles listing and working with receipts
	 */
	private ReceiptsPane		m_ReceiptsPane;

	/**
	 * the confirmation pane, which handles display of a payment confirmation
	 */
	private ConfirmPane		m_ConfirmPane;

	/**
	 * the validate pane, which handles picking and revalidating a voucher
	 */
	private ValidatePane		m_ValidatePane;

	/**
	 * the split pane, which handles picking and splitting a voucher
	 */
	private SplitPane			m_SplitPane;

	/**
	 * the merge pane, which handles picking and merging vouchers together
	 */
	private MergePane			m_MergePane;

	/**
	 * the payment pane, which handles making a voucher payment
	 */
	private PaymentPane		m_PaymentPane;

	/**
	 * the tokens pane, which handles purchasing tokens
	 */
	private TokensPane		m_TokensPane;

	/**
	 * the prices pane, which handles displaying current asset prices
	 */
	private PricesPane		m_PricesPane;

	/**
	 * the incoming pane, which handles incoming voucher payments
	 */
	private IncomingPane		m_IncomingPane;

	/**
	 * the new receipt pane, which handles incoming receipts
	 */
	private NewReceiptPane	m_NewReceiptPane;

	/**
	 * the unclaimed payment pane, which handles outgoing payments not picked up
	 */
	private UnclaimedPane	m_UnclaimedPane;

	/**
	 * the expired voucher pane, which handles deletion of expired vouchers
	 */
	private ExpiredPane		m_ExpiredPane;

	/**
	 * whether or not we should shift to HousekeepingPane immediately upon the
	 * DHT's records reply
	 */
	private boolean			m_PostponeHKP;

	/**
	 * whether to go to overview (main menu) after marking a record read on DHT
	 */
	private boolean			m_GoMainOnMark;

	/**
	 * the pane we were on when we had to buy tokens parenthetically
	 */
	private JPanel				m_PendingTokensPane;

	// END data members
	
	// BEGIN methods
	/**
	 * constructor
	 * @param plugin the containing plugin
	 */
	public WTabManager(WalletClient plugin) {
		super(plugin);

		m_LoginPane = new LoginPane(m_Plugin, this);
		m_HomePane = new HomePane(m_Plugin, this);
		m_MenuPane = new MenuPane(m_Plugin, m_HomePane, this);
		m_HomePane.setWorkPane(m_MenuPane);
	}

	/**
	 * state transition method, called by threads that remove states from queue
	 * @param state the state to transition to
	 */
	public void doState(int state) {
		//Log.debug("WTabManager state transition to " + state);
		WorkPane wp = getWorkPane();
		switch (state) {
			case M_LOGIN1:
				removeAll();
				if (m_LoginPane == null) {
					// this could happen if we logged out and now want to log back
					// in, or if we just created or closed a safe, or recovered one

					m_HomePane = null;
					m_NewSafePane = null;
					m_CloseSafePane = null;
					m_RecoveryPane = null;
					m_LoginPane = new LoginPane(m_Plugin, this);
				}
				m_LoginPane.getPublisherList();
				add(m_LoginPane);
				m_LoginPane.initPhase1();
				break;

			case M_LOGIN2:
				// in this case the object (a LoginPane) was already created, and
				// the challenge characters already recorded

				if (m_LoginPane == null) {
					Log.error("Should already have LoginPane at phase 2");
					break;
				}
				m_LoginPane.initPhase2();
				break;

			case M_NEW_SAFE1:
				if (m_NewSafePane != null) {
					// in this case the object (a NewSafePane) was already created
					m_NewSafePane.initPhase1();
				}
				else {
					// get rid of previous pane and tab containing it
					if (m_LoginPane != null) {
						m_LoginPane.removeAll();
						remove(m_LoginPane);
						m_LoginPane = null;
					}
					else if (m_HomePane != null) {
						remove(m_HomePane);
					}

					// make a new safe pane and recreate the tab using it
					removeAll();
					m_NewSafePane = new NewSafePane(m_Plugin, false, this);
					add(m_NewSafePane);
					m_NewSafePane.initPhase1();
				}
				break;

			case M_NEW_SAFE2:
				// in this case the object (a NewSafePane) was already created
				if (m_NewSafePane == null) {
					Log.error("Should already have NewSafePane at phase 2");
				}
				else {
					m_NewSafePane.initPhase2();
				}
				break;

			case M_NEW_SAFE3:
				// in this case the object (a NewSafePane) was already created
				if (m_NewSafePane == null) {
					Log.error("Should already have NewSafePane at phase 3");
				}
				else {
					m_NewSafePane.initPhase3();
				}
				break;

			case M_NEW_SAFE4:
				// in this case the object (a NewSafePane) was already created
				if (m_NewSafePane == null) {
					Log.error("Should already have NewSafePane at phase 4");
				}
				else {
					m_NewSafePane.initPhase4();
				}
				break;

			case M_HOME_INIT:
				// get rid of login pane and tab containing it
				if (m_LoginPane != null) {
					m_LoginPane.removeAll();
					m_LoginPane = null;
				}
				m_PostponeHKP = false;

				// make a default home pane and recreate the tab using it
				if (m_HomePane == null) {
					m_HomePane = new HomePane(m_Plugin, this);
					m_MenuPane = new MenuPane(m_Plugin, m_HomePane, this);

					/* Tell the home panes about the new HomePane.  This is needed
					 * because they utilize certain data structures from the wallet
					 * tab HomePane, e.g. the VoucherTableModel.
					 */
					XHomePane xhp = m_Plugin.getExchangeTabManager().getHomePane();
					if (xhp != null) {
						xhp.setWalletHome(m_HomePane);
					}
					MHomePane mhp
						= m_Plugin.getMarketplaceTabManager().getHomePane();
					if (mhp != null) {
						mhp.setWalletHome(m_HomePane);
					}
				}
	 			removeAll();
				add(m_HomePane);

				// prepare all the WorkPane subclass objects
				m_HousekeepingPane
					= new HousekeepingPane(m_Plugin, m_HomePane, this);
				m_UnclaimedPane = new UnclaimedPane(m_Plugin, m_HomePane, this);
				m_NewReceiptPane = new NewReceiptPane(m_Plugin, m_HomePane, this);
				m_ReceiptsPane = new ReceiptsPane(m_Plugin, m_HomePane, this);
				m_ConfirmPane = new ConfirmPane(m_Plugin, m_HomePane, this);
				m_ValidatePane = new ValidatePane(m_Plugin, m_HomePane, this);
				m_SplitPane = new SplitPane(m_Plugin, m_HomePane, this);
				m_MergePane = new MergePane(m_Plugin, m_HomePane, this);
				m_PaymentPane = new PaymentPane(m_Plugin, m_HomePane, this);
				m_TokensPane = new TokensPane(m_Plugin, m_HomePane, this);
				m_PricesPane = new PricesPane(m_Plugin, m_HomePane, this);
				m_ExpiredPane = new ExpiredPane(m_Plugin, m_HomePane, this);
				m_IncomingPane = new IncomingPane(m_Plugin, m_HomePane, this);

				// Obtain the network config from the OFS.  This call also gets us
				// the VP's pubkey, just as soon as we find out where the PKS is.

				m_HomePane.setCursor(
										Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
				m_HomePane.updateTitleText();
				m_LoginInit = true;
				JProgressBar pb = new JProgressBar();
				pb.setIndeterminate(true);
				pb.setString("Querying network config, wallet contents, "
								+ "and housekeeping items");
				pb.setStringPainted(true);
				pb.setFont(m_HomePane.M_DataFont);
				pb.setOpaque(true);
				pb.setBackground(Color.LIGHT_GRAY);
				m_HomePane.setWorkPane(m_MenuPane);
				m_MenuPane.disableMenu();
				// NB: re-enable happens on M_WORK_DHT_REGEN
				JPanel statBar = m_Plugin.getStatusBar();
				statBar.removeAll();
				statBar.add(pb);
				statBar.revalidate();
				statBar.repaint();
				retrieveNetworkConfig();
				break;

			case M_HOME_VOUCH_REGEN:
				m_HomePane.rebuildVoucherList();
				if (wp != null) {
					wp.buildValueHeader();
					m_HomePane.revalidate();
					m_HomePane.repaint();
				}
				break;

			case M_HOME_TOKEN_REGEN:
				if (wp != null) {
					wp.buildValueHeader();
					m_HomePane.revalidate();
					m_HomePane.repaint();
				}
				break;

			case M_WORK_DHT_REGEN:
				/* have the housekeeping pane record and process any updates
				 * (note there might not be any if we're just marking a record read)
				 */
				if (m_LoginInit) {
					// only go to housekeeping if we have some to do: DHT items
					// or expired payments

					PaymentTableModel ptm = m_HousekeepingPane.getPaymentModel();
					Hashtable<XMLPayment, String> pmts = ptm.getPaymentList();
					boolean expPay = false;
					for (XMLPayment pmt : pmts.keySet()) {
						if (pmt.isExpired()) {
							expPay = true;
							break;
						}
					}
					ConcurrentHashMap<Integer, String> dhtRawRecs
						= m_Plugin.getDHTListener().getDHTrecs();
					if (!dhtRawRecs.isEmpty() || expPay) {
						m_HousekeepingPane.updateDHTrecords();
						m_HousekeepingPane.initProcessing();
						m_HousekeepingPane.doHousekeeping();
						m_LoginInit = false;
					}
					else {
						m_Plugin.setDefaultStatusBar();
						m_LoginInit = false;
						m_HomePane.setWorkPane(m_MenuPane);
					}
					m_MenuPane.enableMenu();
				}
				else {
					/* if we're postponing Housekeeping display, decode the DHT
					 * records on the current screen display
					 */
					if (!m_PostponeHKP && !m_GoMainOnMark) {
						// normal situation when Refresh pushed
						m_HousekeepingPane.updateDHTrecords();
						m_HousekeepingPane.doHousekeeping();
					}
					else {
						m_HousekeepingPane.updateDHTrecords();
					}
					if (m_PostponeHKP) {
						m_PostponeHKP = false;
						m_HousekeepingPane.doHousekeeping();
					}

					// if we're supposed to return to main menu, go there
					if (m_GoMainOnMark) {
						m_GoMainOnMark = false;
						m_HomePane.setWorkPane(m_MenuPane);

						/* Additionally, we need to check with the SVX tab manager's
						 * TradePane/AcceptPane/OffersPane/SellPane to see whether
						 * we need to continue an offer placement/accept after
						 * completing an escrow voucher payment.  This is indicated
						 * by having a flag set in these UI objects.
						 *
						 * Similarly, for an SVM tab entity the tab manager may have
						 * a designated recipient for confirmEscrowFunding() calls.
						 */
						XTabManager xtm = m_Plugin.getExchangeTabManager();
						TradePane tp = xtm.getTradePane();
						OffersPane op = xtm.getOffersPane();
						AcceptPane ap = xtm.getAcceptPane();
						SellPane sp = xtm.getSellPane();
						// (this MTabManager may be a derived class found in a plugin)
						MTabManager mtm = m_Plugin.getMarketplaceTabManager();
						org.silentvault.client.ui.svm.WorkPane escHandler
							= mtm.getEscrowHandler();
						if (tp != null && tp.doingEscrow()) {
							tp.confirmEscrowFunding();
						}
						else if (op != null && op.doingEscrow()) {
							op.confirmEscrowFunding();
						}
						else if (ap != null && ap.doingEscrow()) {
							ap.confirmEscrowFunding();
						}
						else if (sp != null && sp.doingEscrow()) {
							sp.confirmEscrowFunding();
						}
						else if (escHandler != null && escHandler.doingEscrow()) {
							escHandler.confirmEscrowFunding();
						}
					}
				}
				break;

			case M_WORK_RECEIPT_REGEN:
				// rebuild the receipts display in work pane
				m_HomePane.setCursor(null);
				m_ReceiptsPane.rebuildReceiptList();
				m_HomePane.setWorkPane(m_ReceiptsPane);
				break;

			case M_WORK_PAYMENT_REGEN:
				// rebuild the pending payments list in housekeeping pane
				m_HomePane.setCursor(null);
				if (wp != null) {
					m_HousekeepingPane.updatePaymentRecords();
					if (m_LoginInit) {
						// lastly, go query the DHT (we do it serially to avoid
						// a race condition between the SDS and DHT responses)

						m_HomePane.setCursor(Cursor.getPredefinedCursor(
																			Cursor.WAIT_CURSOR));
						queryDHTrecords(false);
					}
					else {
						m_HousekeepingPane.doHousekeeping();
					}
					m_HomePane.revalidate();
					m_HomePane.repaint();
				}
				break;

			case M_RECOV1:
				if (m_RecoveryPane != null) {
					// in this case the object (a RecoveryPane) was already created
					m_RecoveryPane.initPhase1();
				}
				else {
					// get rid of login pane and tab containing it
					m_LoginPane.removeAll();
					m_LoginPane = null;

					// make a recover safe pane and recreate the tab using it
					m_RecoveryPane = new RecoveryPane(m_Plugin, this);
					removeAll();
					add(m_RecoveryPane);
					m_RecoveryPane.initPhase1();
				}
				break;

			case M_RECOV2:
				// in this case the object (a RecoveryPane) was already created
				if (m_RecoveryPane == null) {
					Log.error("Should already have RecoveryPane at phase 2");
				}
				else {
					m_RecoveryPane.initPhase2();
				}
				break;

			case M_RECOV3:
				// in this case the object (a RecoveryPane) was already created
				if (m_RecoveryPane == null) {
					Log.error("Should already have RecoveryPane at phase 3");
				}
				else {
					m_RecoveryPane.initPhase3();
				}
				break;

			case M_CLOSE_SAFE1:
				// make a close safe pane and initialize it (doesn't display yet)
				m_CloseSafePane = new CloseSafePane(m_Plugin, this);

				// this call also checks to see if it's okay to close this VS
				if (!m_CloseSafePane.initPhase1(m_Plugin.getLoginSecrets())) {
					// user has to do something else before they can do this
					return;
				}

				// see whether we have value that needs to be transferred out
				if (m_CloseSafePane.needSuccessor()) {
					// The first thing we have to do is create a replacement safe.
					// So make a new safe pane and recreate the tab using it.

					m_NewSafePane = new NewSafePane(m_Plugin, true, this);
					removeAll();
					add(m_NewSafePane);

					// the publisher must remain the same as the closing safe VP
					ArrayList<String> pubList = new ArrayList<String>(1);
					pubList.add(m_Plugin.getLoginSecrets().getPublisher());
					m_NewSafePane.setPubList(pubList);

					// proceed to phase 1 of new safe creation
					m_NewSafePane.initPhase1();
				}
				else {
					// can do it without a successor safe, go on to next step
					doState(M_CLOSE_SAFE2);
					return;
				}
				break;

			case M_CLOSE_SAFE2:
				// proceed to user confirmation of safe closure
				if (m_CloseSafePane != null) {
					// We got here from the final state in NewSafePane, or from the
					// else clause in M_CLOSE_SAFE1 above.

					removeAll();
					add(m_CloseSafePane);
					m_CloseSafePane.initPhase2();
				}
				else {
					Log.error("Impossible state " + state + " without CloseSafePane "
								+ "object");
				}
				break;

			case M_LOGOUT_SAFE:
				// user was auto-logged out, show message screen leading to login
				removeAll();
				m_HomePane = null;
				m_NewSafePane = null;
				m_CloseSafePane = null;
				m_LoginPane = new LoginPane(m_Plugin, this);
				add(m_LoginPane);
				m_LoginPane.initPhase0();
				break;

			case M_PREFERENCES:
				// disallow this if we're already in certain states where we must
				// not be interrupted

				switch (m_State) {
					case M_LOGIN2:
					case M_NEW_SAFE2:
					case M_NEW_SAFE3:
					case M_NEW_SAFE4:
					case M_RECOV3:
					case M_CLOSE_SAFE1:
					case M_CLOSE_SAFE2:
					case M_PAYMENTS1:
					case M_INCOMING_PAYMENT:
					case M_INCOMING_RECEIPT:
					case M_UNCLAIMED:
					case M_EXPIRED:
					case M_CONFIRMATION:
					case M_PURCHASE_TOKENS:
						showError("Cannot switch to preferences from this screen");
						return;
				}

				// make the preferences UI the displayed frame
				JComponent prefPane = m_Plugin.getUserPreferences().getGUI();
				if (prefPane != null) {
					JPanel prefPanel = new JPanel();
					prefPanel.setLayout(new GridLayout(1, 1));
					prefPanel.setOpaque(true);
					prefPanel.add(prefPane);
					removeAll();
					add(prefPanel);
				}
				else {
					Log.error("Missing preferences UI component");
				}
				break;

			case M_HOME_RETURN:
				// return to the home pane after preferences
				if (m_HomePane != null) {
					removeAll();
					add(m_HomePane);
					// for some reason, need to reset the border on the home pane
					String barTxt = "Wallet ID: "
										+ m_Plugin.getLoginSecrets().getVSnumber() + " ";
					m_HomePane.setBorder(BorderFactory.createTitledBorder(barTxt));

					// update the voucher table header in case currency pref changed
					String prefCurrency
						= m_Plugin.getUserPreferences().getValueCurrency();
					VoucherTableModel vtm = m_HomePane.getVoucherModel();
					vtm.setEquivHeader(prefCurrency);
					m_HomePane.configTableColumns(m_HomePane.getVoucherTable());
					wp.buildValueHeader();
				}
				else {
					Log.error("No home pane to return to after preferences!");
				}
				break;

			case M_DISCONNECTED:
				// show a screen indicating disconnection
				removeAll();
				JPanel discPanel = new JPanel();
				discPanel.setOpaque(true);
				GridBagLayout gbl = new GridBagLayout();
				GridBagConstraints gbc = new GridBagConstraints();
				discPanel.setLayout(gbl);
				gbc.weighty = 1.0;
				gbc.gridwidth = GridBagConstraints.REMAINDER;
				gbc.gridheight = 5;
				gbc.anchor = GridBagConstraints.CENTER;
				discPanel.setBorder(BorderFactory.createTitledBorder(
																"Disconnected from network"));
				String discTxt = "\nYou are disconnected from the gateway.\n\n"
										+ "In order to log in, you will need to "
										+ "connect first.\n";
				JTextArea disc = new JTextArea(discTxt, 5, 40);
				disc.setLineWrap(true);
				disc.setWrapStyleWord(true);
				disc.setEditable(false);
				disc.setBorder(null);
				disc.setBackground(m_Plugin.m_LabelBackColor);
				discPanel.add(Box.createVerticalGlue(), gbc);
				discPanel.add(Box.createVerticalStrut(50), gbc);
				discPanel.add(disc, gbc);
				discPanel.add(Box.createVerticalStrut(50), gbc);
				discPanel.add(Box.createVerticalGlue(), gbc);
				add(discPanel);
				break;

			case M_PAYMENTS1:
				// check token sufficiency
				TokenTableModel ttm = m_HomePane.getTokenModel();
				VoucherTableModel vtm = m_HomePane.getVoucherModel();
				if (ttm.getAvailableCount(vtm.getSelectedIssuer())
					 < VoucherRequest.M_PaymentCost)
				{
					setPendingTokens(m_PaymentPane);
					m_HomePane.setWorkPane(m_TokensPane);
				}
				else {
					m_HomePane.setWorkPane(m_PaymentPane);
				}
				break;

			case M_RECEIPT_LIST:
				m_HomePane.setWorkPane(m_ReceiptsPane);
				break;

			case M_VALIDATE:
				// check token sufficiency
				ttm = m_HomePane.getTokenModel();
				vtm = m_HomePane.getVoucherModel();
				if (ttm.getAvailableCount(vtm.getSelectedIssuer())
					 < VoucherRequest.M_ValidateCost)
				{
					setPendingTokens(m_ValidatePane);
					m_HomePane.setWorkPane(m_TokensPane);
				}
				else {
					m_HomePane.setWorkPane(m_ValidatePane);
				}
				break;

			case M_SPLIT:
				// check token sufficiency
				ttm = m_HomePane.getTokenModel();
				vtm = m_HomePane.getVoucherModel();
				if (ttm.getAvailableCount(vtm.getSelectedIssuer())
					 < VoucherRequest.M_SplitCost)
				{
					setPendingTokens(m_SplitPane);
					m_HomePane.setWorkPane(m_TokensPane);
				}
				else {
					m_HomePane.setWorkPane(m_SplitPane);
				}
				break;

			case M_MERGE:
				// check token sufficiency
				ttm = m_HomePane.getTokenModel();
				vtm = m_HomePane.getVoucherModel();
				if (ttm.getAvailableCount(vtm.getSelectedIssuer())
					 < VoucherRequest.M_MergeCost)
				{
					setPendingTokens(m_MergePane);
					m_HomePane.setWorkPane(m_TokensPane);
				}
				else {
					m_HomePane.setWorkPane(m_MergePane);
				}
				break;

			case M_PRICES:
				// go get latest prices
				AssetPriceCache apc = m_Plugin.getPricingModule();
				if (apc != null) {
					apc.run();
				}
				m_HomePane.setWorkPane(m_PricesPane);
				break;

			case M_GO_MAIN_MENU:
				m_HomePane.setWorkPane(m_MenuPane);
				break;

			case M_INCOMING_PAYMENT:
				// NB: because the logic is complex, token sufficiency is handled
				//		 in the sub-pane
				m_HomePane.setWorkPane(m_IncomingPane);
				break;

			case M_INCOMING_RECEIPT:
				// check token sufficiency
				ttm = m_HomePane.getTokenModel();
				vtm = m_HomePane.getVoucherModel();
				if (ttm.getAvailableCount(vtm.getSelectedIssuer())
					 < VoucherRequest.M_ReceiptCost)
				{
					setPendingTokens(m_NewReceiptPane);
					m_HomePane.setWorkPane(m_TokensPane);
				}
				else {
					m_HomePane.setWorkPane(m_NewReceiptPane);
				}
				break;

			case M_UNCLAIMED:
				m_HomePane.setWorkPane(m_UnclaimedPane);
				break;

			case M_EXPIRED:
				m_HomePane.setWorkPane(m_ExpiredPane);
				break;

			case M_HOUSEKEEPING_RETURN:
				m_HousekeepingPane.doHousekeeping();
				break;

			case M_CONFIRMATION:
				m_HomePane.setWorkPane(m_ConfirmPane);
				break;

			case M_PURCHASE_TOKENS:
				m_HomePane.setWorkPane(m_TokensPane);
				break;

			case M_PENDING_TOKENS:
				// return to whichever pane we were on when tokens needed purchasing
				if (m_PendingTokensPane != null) {
					if (m_PendingTokensPane instanceof CloseSafePane) {
						m_PendingTokensPane = null;
						doState(M_CLOSE_SAFE1);
					}
					else {
						m_HomePane.setWorkPane((WorkPane) m_PendingTokensPane);
						m_PendingTokensPane = null;
					}
				}
				else {
					// provide a default back to the main menu
					m_HomePane.setWorkPane(m_MenuPane);
				}
				break;

			case M_TUTORIALS:
				// get the desktop browser for this system and try to open the URI
				URI tutURI = null;
				boolean needManual = false;
				try {
					tutURI = new URI("https://www.voucher-safe.com");
				}
				catch (URISyntaxException use) {
					Log.error("Bad URI for tutorials");
					needManual = true;
				}
				Desktop browser = null;
				try {
					browser = Desktop.getDesktop();
				}
				catch (UnsupportedOperationException uoe) {
					needManual = true;
				}
				finally {
					if (!needManual) {
						if (!browser.isSupported(Desktop.Action.BROWSE)) {
							needManual = true;
						}
					}
				}
				if (needManual) {
					// just put up the URL in a popup window with instructions
					JFrame tutPop = new JFrame("VSC Tutorials");
					tutPop.setDefaultCloseOperation(
															WindowConstants.DISPOSE_ON_CLOSE);
					JEditorPane tutViewer = new JEditorPane("text/html", null);
					tutViewer.setBackground(Color.LIGHT_GRAY);
					StringBuilder content = new StringBuilder(512);
					content.append("<html><body bgcolor=\"silver\">");
					content.append("<center><br/><br/><p>");
					content.append("For tutorials help, paste the following URL ");
					content.append("into your browser:</p><br/><br/>");
					content.append("https://www.voucher-safe.com<br/>");
					content.append("</center></body></html>");
					tutViewer.setText(content.toString());
					tutPop.add(tutViewer);
					tutPop.setSize(new Dimension(400, 200));
					tutPop.setLocationRelativeTo(this);
					tutPop.setVisible(true);
					m_Plugin.registerPopup(tutPop);
				}
				else {
					try {
						// actually invoke the default browser
						browser.browse(tutURI);
					}
					catch (IOException ioe) {
						Log.error("Could not invoke browser on tutorials page", ioe);
						showError("Could not display tutorials",
									"Error opening default browser",
								"Point your browser to: https://www.voucher-safe.com");
					}
				}
				break;

			default:
				Log.error("Unsupported UI state " + state);
				break;
		}
		if (state != M_TUTORIALS) {
			m_State = state;
			revalidate();
			repaint();
		}
	}

	/**
	 * set the pane to go back to once a token purchase is completed
	 * @param pane the pane
	 */
	public void setPendingTokens(JPanel pane) { m_PendingTokensPane = pane; }


	// get methods
	/**
	 * obtain the login pane
	 * @return the pane used for logins
	 */
	public LoginPane getLoginPane() { return m_LoginPane; }

	/**
	 * obtain the home pane
	 * @return the pane used for home safe display
	 */
	public HomePane getHomePane() { return m_HomePane; }

	/**
	 * obtain the work pane (a component in the home pane which contains the
	 * WorkPane subclass)
	 * @return the pane used for work area display
	 */
	public WorkPane getWorkPane() {
		if (m_HomePane == null) {
			return null;
		}
		return m_HomePane.getWorkPane();
	}

	/**
	 * get the housekeeping pane, which handles items retrieved from the DHT
	 * @return the pane used for housekeeping
	 */
	public HousekeepingPane	getHousekeepingPane() { return m_HousekeepingPane; }

	/**
	 * get the unclaimed pane, which handles pending payments (expired or not)
	 * @return the pane used for unclaimed payments
	 */
	public UnclaimedPane getUnclaimedPane() { return m_UnclaimedPane; }

	/**
	 * get the receipt pane, which handles display of a single incoming receipt
	 * @return the pane used for new receipts downloaded off the DHT
	 */
	public NewReceiptPane getNewReceiptPane() { return m_NewReceiptPane; }

	/**
	 * get the main menu pane, which is the home state after login
	 * @return the pane used for the main menu
	 */
	public MenuPane getMenuPane() { return m_MenuPane; }

	/**
	 * get the receipts pane, which handles listing and working with receipts
	 * @return the pane used for receipt lists
	 */
	public ReceiptsPane getReceiptsPane() { return m_ReceiptsPane; }

	/**
	 * get the confirmation pane, which handles display of a payment confirmation
	 * @return the pane used for payment confirmations
	 */
	public ConfirmPane getConfirmPane() { return m_ConfirmPane; }

	/**
	 * get the validate pane, which handles picking and revalidating a voucher
	 * @return the pane used for voucher validation
	 */
	public ValidatePane getValidatePane() { return m_ValidatePane; }

	/**
	 * get the split pane, which handles picking and splitting a voucher
	 * @return the pane used for voucher splits
	 */
	public SplitPane getSplitPane() { return m_SplitPane; }

	/**
	 * get the merge pane, which handles picking and merging vouchers together
	 * @return the pane used for voucher merges
	 */
	public MergePane getMergePane() { return m_MergePane; }

	/**
	 * get the payment pane, which handles making a voucher payment
	 * @return the pane used for voucher payments
	 */
	public PaymentPane getPaymentPane() { return m_PaymentPane; }

	/**
	 * get the tokens pane, which handles purchasing tokens
	 * @return the pane used for token purchases
	 */
	public TokensPane getTokensPane() { return m_TokensPane; }

	/**
	 * get the prices pane, which handles displaying current asset prices
	 * @return the pane used for displaying price info
	 */
	public PricesPane getPricesPane() { return m_PricesPane; }

	/**
	 * get the incoming pane, which handles incoming voucher payments
	 * @return the pane used for processing incoming payments
	 */
	public IncomingPane getIncomingPane() { return m_IncomingPane; }

	/**
	 * get the expired voucher pane, which handles deleting expired vouchers
	 * @return the pane used to delete expired vouchers
	 */
	public ExpiredPane getExpiredPane() { return m_ExpiredPane; }

	/**
	 * obtain the new safe pane
	 * @return the pane used for new safe creation
	 */
	public NewSafePane getNewSafePane() { return m_NewSafePane; }

	/**
	 * obtain the recovery pane
	 * @return the pane used for safe data recovery
	 */
	public RecoveryPane getRecoveryPane() { return m_RecoveryPane; }

	/**
	 * obtain the close safe pane
	 * @return the pane used for closing a safe
	 */
	public CloseSafePane getCloseSafePane() { return m_CloseSafePane; }

	/**
	 * obtain the pane we were on when a token purchase began
	 * @return the pane (a WorkPane sub-class), or null if we used menu item
	 */
	public JPanel getPendingTokensPane() { return m_PendingTokensPane; }


	// generic query methods
	/**
	 * method to query the OFS for the network config, following login.
	 * NB: this method blocks waiting for the network config reply.
	 */
	private void retrieveNetworkConfig() {
		final XMPPConnection conn = m_Plugin.getOFSConnection();
		VscState stateData = m_Plugin.getVscState();

		// build the message to send
		OFSMessage netDisco = new OFSMessage();
		netDisco.setType(IQ.Type.GET);
		netDisco.setFrom(conn.getUser());
		netDisco.setTo(conn.getServiceName());
		netDisco.setPacketID("disco_" + stateData.getNextOFSid());
		netDisco.setOpcode("REQ_discover_net");
		netDisco.setVoucher_publisher(m_Plugin.getLoginSecrets().getPublisher());

		// create a packet filter which can look for the reply
		PacketIDFilter pFilter = new PacketIDFilter(netDisco.getPacketID());
		PacketCollector repColl = conn.createPacketCollector(pFilter);

		// send it and record outgoing packet
		conn.sendPacket(netDisco);
		stateData.setLastActivity(netDisco);
		//Log.debug("Sent disco request");

		// await synchronous reply
		OFSMessage rep = null;
		IQ reply = (IQ) repColl.nextResult(m_Plugin.getTimeout());
		if (reply == null) {
			// got timeout; build error
			rep = new OFSMessage();
			rep.setOpcode("REP_network_details");
			rep.setPacketID(netDisco.getPacketID());
			rep.setType(IQ.Type.ERROR);
			rep.setErrcode(java.net.HttpURLConnection.HTTP_GATEWAY_TIMEOUT);
			rep.setErrmsg("timed out");
		}
		else {
			rep = (OFSMessage) reply;
		}

		// invoke callback with result
		m_Plugin.getOFSListener().processPacket(rep);
	}

	/**
	 * method to query the SDS for the list of vouchers and tokens in the safe.
	 * (i.e. to do both at once in parallel)  If this method returns failure,
	 * the caller should log out of the wallet.
	 * @param forceSync if true, wait for replies
	 * @return true if it worked, false otherwise (usually SDS connect failure)
	 */
	public boolean querySafeContents(boolean forceSync) {
		VscState state = m_Plugin.getVscState();
		if (!state.isLoggedIn()) {
			Log.error("Cannot list wallet contents while not logged in!");
			showError("cannot list wallet contents while not logged in");
			return false;
		}
		
		// open a connection to the SDS if we don't have one
		XMPPConnection sdsConn = m_Plugin.getSDSListener().getSDSConnection();
		if (sdsConn == null) {
			sdsConn = m_Plugin.openSDS();
		}
		if (sdsConn == null) {
			// can't do anything
			Log.error("Cannot list wallet contents due to no SDS connection");
			showError("error accessing your wallet",
						"cannot obtain wallet contents due to no SDS connection",
						"please try again later");

			// fail -- caller should force a logout
			return false;
		}

		if (forceSync) {
			m_HomePane.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
		}
		queryTokens(sdsConn, forceSync);
		queryVouchers(sdsConn, forceSync);
		return true;
	}

	/**
	 * helper method to send REQ_list_vouchers to SDS
	 * @param sdsConn the active SDS connection, or null to obtain our own
	 * @param forceSync if true, wait for replies
	 */
	public void queryVouchers(XMPPConnection sdsConn, boolean forceSync) {
		if (sdsConn == null)  {
			sdsConn = m_Plugin.getSDSListener().getSDSConnection();
			if (sdsConn == null) {
				sdsConn = m_Plugin.openSDS();
				if (sdsConn == null) {
					return;
				}
			}
		}
		VsSecrets secrets = m_Plugin.getLoginSecrets();
		VscState state = m_Plugin.getVscState();

		// build message to send for obtaining vouchers
		SDSMessage getVouchers = new SDSMessage();
		getVouchers.setType(IQ.Type.GET);
		getVouchers.setFrom(sdsConn.getUser());
		getVouchers.setTo(sdsConn.getServiceName());
		getVouchers.setPacketID("vlist_" + state.getNextSDSid());
		getVouchers.setOpcode("REQ_list_vouchers");
		String vIdx = new String(secrets.getVoucherIndex());
		getVouchers.setFolder_index(vIdx);
		String rCap = new String(secrets.getVoucherReadCap());
		getVouchers.setReadcap(rCap);

		// create a packet filter which can look for the reply
		PacketIDFilter pFilter = new PacketIDFilter(getVouchers.getPacketID());
		PacketCollector repColl = null;
		if (forceSync) {
			repColl = sdsConn.createPacketCollector(pFilter);
		}
		else {
			m_Plugin.getSDSListener().addFilter(getVouchers.getPacketID(),
															pFilter);
		}

		// send request to the SDS
		sdsConn.sendPacket(getVouchers);
		//Log.debug(getVouchers.getOpcode() + " sent with index: " + vIdx);

		// record outgoing packet
		state.setLastActivity(getVouchers);

		// wait for reply if asked to do so
		if (forceSync) {
			// wait for reply or timeout
			SDSMessage rep = null;
			IQ reply = (IQ) repColl.nextResult(m_Plugin.getTimeout());
			if (reply == null) {
				// got timeout; build error
				rep = new SDSMessage();
				rep.setOpcode("REP_voucher_list");
				rep.setFrom(getVouchers.getTo());
				rep.setTo(getVouchers.getFrom());
				rep.setPacketID(getVouchers.getPacketID());
				rep.setType(IQ.Type.ERROR);
				rep.setErrcode(java.net.HttpURLConnection.HTTP_GATEWAY_TIMEOUT);
				rep.setErrmsg("timed out");
			}
			else {
				rep = (SDSMessage) reply;
			}

			// invoke callback with result
			m_Plugin.getSDSListener().processPacket(rep);
		}
	}

	/**
	 * helper method to send REQ_list_tokens to SDS
	 * @param sdsConn the active SDS connection, or null to obtain our own
	 * @param forceSync if true, wait for replies
	 */
	public void queryTokens(XMPPConnection sdsConn, boolean forceSync) {
		if (sdsConn == null) {
			sdsConn = m_Plugin.getSDSListener().getSDSConnection();
			if (sdsConn == null) {
				sdsConn = m_Plugin.openSDS();
				if (sdsConn == null) {
					return;
				}
			}
		}
		VsSecrets secrets = m_Plugin.getLoginSecrets();
		VscState state = m_Plugin.getVscState();

		// build message to send for obtaining tokens
		SDSMessage getTokens = new SDSMessage();
		getTokens.setType(IQ.Type.GET);
		getTokens.setFrom(sdsConn.getUser());
		getTokens.setTo(sdsConn.getServiceName());
		getTokens.setPacketID("tlist_" + state.getNextSDSid());
		getTokens.setOpcode("REQ_list_tokens");
		String vIdx = new String(secrets.getVoucherIndex());
		getTokens.setFolder_index(vIdx);
		String rCap = new String(secrets.getVoucherReadCap());
		getTokens.setReadcap(rCap);

		// create a packet filter which can look for the reply
		PacketIDFilter pFilter = new PacketIDFilter(getTokens.getPacketID());
		PacketCollector repColl = null;
		if (forceSync) {
			repColl = sdsConn.createPacketCollector(pFilter);
		}
		else {
			m_Plugin.getSDSListener().addFilter(getTokens.getPacketID(),
															pFilter);
		}

		// send request to the SDS
		sdsConn.sendPacket(getTokens);
		//Log.debug(getTokens.getOpcode() + " sent with index: " + vIdx);

		// record outgoing packet
		state.setLastActivity(getTokens);

		// wait for reply if asked to do so
		if (forceSync) {
			// wait for reply or timeout
			SDSMessage rep = null;
			IQ reply = (IQ) repColl.nextResult(m_Plugin.getTimeout());
			if (reply == null) {
				// got timeout; build error
				rep = new SDSMessage();
				rep.setOpcode("REP_token_list");
				rep.setFrom(getTokens.getTo());
				rep.setTo(getTokens.getFrom());
				rep.setPacketID(getTokens.getPacketID());
				rep.setType(IQ.Type.ERROR);
				rep.setErrcode(java.net.HttpURLConnection.HTTP_GATEWAY_TIMEOUT);
				rep.setErrmsg("timed out");
			}
			else {
				rep = (SDSMessage) reply;
			}

			// invoke callback with result
			m_Plugin.getSDSListener().processPacket(rep);
		}
	}

	/**
	 * method to query the DHT for the list of unmarked records for this safe
	 * (always synchronous)
	 * @param noHKP if set, do not switch to housekeeping until replies processed
	 */
	public void queryDHTrecords(boolean noHKP) {
		VsSecrets secrets = m_Plugin.getLoginSecrets();
		VscState state = m_Plugin.getVscState();
		if (!state.isLoggedIn()) {
			m_HomePane.setCursor(null);
			Log.error("Cannot list DHT records while not logged in!");
			showError("cannot list DHT records", "not logged in", "log in first");
			return;
		}
		
		// open a connection to the DHT if we don't have one
		XMPPConnection dhtConn = m_Plugin.getDHTListener().getDHTConnection();
		if (dhtConn == null) {
			dhtConn = m_Plugin.openDHT();
		}
		if (dhtConn == null) {
			// can't do anything
			m_HomePane.setCursor(null);
			Log.error("Cannot list DHT records due to no DHT connection");
			showError("cannot list DHT records", "no DHT connection");
			return;
		}
		m_PostponeHKP = noHKP;

		// build message to send for obtaining records not yet marked with hash
		DHTMessage getRecs = new DHTMessage();
		getRecs.setType(IQ.Type.GET);
		getRecs.setFrom(dhtConn.getUser());
		getRecs.setTo(dhtConn.getServiceName());
		getRecs.setPacketID("get_" + state.getNextDHTid());
		getRecs.setOpcode("REQ_get_records");

		// our hashkey is hash(VS#)
		MessageDigest md = null;
		try {
			md = MessageDigest.getInstance("SHA-1");
		}
		catch (NoSuchAlgorithmException nsae) {
			// impossible
			Log.error("Can't build SHA-1 hash", nsae);
		}
		md.update(secrets.getVSnumber().getBytes());
		String hashKey = StringUtils.encodeBase64(md.digest(), false);
		getRecs.setRecipient_hash(hashKey);

		// our secret hash is: hash(VS#+PIN+passphrase)
		getRecs.setSecret_hash(new String(secrets.getDHTMarkHash()));

		// create a packet filter (async) which can look for the reply
		PacketIDFilter pFilter = new PacketIDFilter(getRecs.getPacketID());
		PacketCollector repColl = dhtConn.createPacketCollector(pFilter);

		// send request to the DHT
		dhtConn.sendPacket(getRecs);
		//Log.debug(getRecs.getOpcode() + " sent with hashkey: " + hashKey);

		// record outgoing packet
		state.setLastActivity(getRecs);

		// wait for reply or timeout
		DHTMessage rep = null;
		IQ reply = (IQ) repColl.nextResult(m_Plugin.getTimeout());
		if (reply == null) {
			// got timeout; build error
			rep = new DHTMessage();
			rep.setOpcode("REP_records");
			rep.setFrom(getRecs.getTo());
			rep.setTo(getRecs.getFrom());
			rep.setPacketID(getRecs.getPacketID());
			rep.setType(IQ.Type.ERROR);
			rep.setErrcode(java.net.HttpURLConnection.HTTP_GATEWAY_TIMEOUT);
			rep.setErrmsg("timed out");
		}
		else {
			rep = (DHTMessage) reply;
		}

		// invoke callback with result
		m_Plugin.getDHTListener().processPacket(rep);
	}

	/**
	 * method to mark a record read on the DHT
	 * @param recId the id of the record to be marked
	 * @param goMain if true, go to the main menu on return, not housekeeping
	 */
	public void markReadOnDHT(Integer recId, boolean goMain) {
		if (recId == null) {
			Log.error("Null DHT record specified for marking");
			return;
		}
		VsSecrets secrets = m_Plugin.getLoginSecrets();
		VscState state = m_Plugin.getVscState();
		if (!state.isLoggedIn()) {
			Log.error("Cannot mark DHT records while not logged in!");
			showError("cannot mark DHT records", "not logged in", "log in first");
			return;
		}
		m_GoMainOnMark = goMain;
		
		// open a connection to the DHT if we don't have one
		DHTListener dhtl = m_Plugin.getDHTListener();
		XMPPConnection dhtConn = dhtl.getDHTConnection();
		if (dhtConn == null) {
			dhtConn = m_Plugin.openDHT();
		}
		if (dhtConn == null) {
			// can't do anything
			Log.error("Cannot mark DHT records due to no DHT connection");
			showError("cannot mark DHT records", "no DHT connection");
			return;
		}

		// build message to send for marking records
		DHTMessage markRec = new DHTMessage();
		markRec.setType(IQ.Type.SET);
		markRec.setFrom(dhtConn.getUser());
		markRec.setTo(dhtConn.getServiceName());
		markRec.setPacketID("mark_" + state.getNextDHTid());
		markRec.setOpcode("REQ_mark_records");
		markRec.markRecord(recId.intValue());

		// our hashkey is hash(VS#)
		MessageDigest md = null;
		try {
			md = MessageDigest.getInstance("SHA-1");
		}
		catch (NoSuchAlgorithmException nsae) {
			// impossible
			Log.error("Can't build SHA-1 hash", nsae);
		}
		md.update(secrets.getVSnumber().getBytes());
		String hashKey = StringUtils.encodeBase64(md.digest(), false);
		markRec.setRecipient_hash(hashKey);

		// our secret hash is: hash(VS#+PIN+passphrase)
		markRec.setSecret_hash(new String(secrets.getDHTMarkHash()));

		// create a packet listener (async) to monitor for reply
		PacketIDFilter pFilter = new PacketIDFilter(markRec.getPacketID());
		m_Plugin.getDHTListener().addFilter(markRec.getPacketID(), pFilter);

		// send request to the DHT
		dhtl.recordMarkReq(markRec);
		dhtConn.sendPacket(markRec);
		//Log.debug(markRec.getOpcode() + " sent with hashkey: " + hashKey);

		// record outgoing packet
		state.setLastActivity(markRec);
	}

	/**
	 * method to deallocate all sub-panes (called by WalletClient destructor)
	 */
	public void deallocatePanes() {
		if (m_LoginPane != null) {
			m_LoginPane.removeAll();
			m_LoginPane = null;
		}
		if (m_NewSafePane != null) {
			m_NewSafePane.removeAll();
			m_NewSafePane = null;
		}
		if (m_RecoveryPane != null) {
			m_RecoveryPane.removeAll();
			m_RecoveryPane = null;
		}
		if (m_CloseSafePane != null) {
			m_CloseSafePane.removeAll();
			m_CloseSafePane = null;
		}
		if (m_HomePane != null) {
			m_HomePane.removeAll();
			m_HomePane = null;
		}
		if (m_HousekeepingPane != null) {
			m_HousekeepingPane.removeAll();
			m_HousekeepingPane = null;
		}
		if (m_UnclaimedPane != null) {
			m_UnclaimedPane.removeAll();
			m_UnclaimedPane = null;
		}
		if (m_NewReceiptPane != null) {
			m_NewReceiptPane.removeAll();
			m_NewReceiptPane = null;
		}
		if (m_MenuPane != null) {
			m_MenuPane.removeAll();
			m_MenuPane = null;
		}
		if (m_ReceiptsPane != null) {
			m_ReceiptsPane.removeAll();
			m_ReceiptsPane = null;
		}
		if (m_ConfirmPane != null) {
			m_ConfirmPane.removeAll();
			m_ConfirmPane = null;
		}
		if (m_ValidatePane != null) {
			m_ValidatePane.removeAll();
			m_ValidatePane = null;
		}
		if (m_SplitPane != null) {
			m_SplitPane.removeAll();
			m_SplitPane = null;
		}
		if (m_MergePane != null) {
			m_MergePane.removeAll();
			m_MergePane = null;
		}
		if (m_PaymentPane != null) {
			m_PaymentPane.removeAll();
			m_PaymentPane = null;
		}
		if (m_TokensPane != null) {
			m_TokensPane.removeAll();
			m_TokensPane = null;
		}
		if (m_PricesPane != null) {
			m_PricesPane.removeAll();
			m_PricesPane = null;
		}
		if (m_ExpiredPane != null) {
			m_ExpiredPane.removeAll();
			m_ExpiredPane = null;
		}
		if (m_IncomingPane != null) {
			m_IncomingPane.removeAll();
			m_IncomingPane = null;
		}
	}

	// END methods
}
